# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import grpc
import pytest
import time
from mercurial import (
    pycompat,
)
from mercurial.node import nullhex as NULL_HEX
from mercurial_testhelpers import (
    as_bytes,
    RepoWrapper,
)
from google.protobuf.timestamp_pb2 import Timestamp
from hgext3rd.heptapod.branch import (
    set_default_gitlab_branch,
    write_gitlab_branches,
)
from hgext3rd.heptapod.branch import (
    write_gitlab_tags,
)
from hgext3rd.heptapod.special_ref import (
    write_special_refs,
)

from hgitaly.oid import (
    blob_oid,
)

from hgitaly.tests.common import (
    make_empty_repo,
    make_tree_shaped_repo,
)

from hgitaly.stub.commit_pb2 import (
    CommitIsAncestorRequest,
    CommitsByMessageRequest,
    CommitLanguagesRequest,
    CommitLanguagesResponse,
    CommitStatsRequest,
    CountCommitsRequest,
    CountDivergingCommitsRequest,
    FindCommitRequest,
    FindCommitsRequest,
    FindAllCommitsRequest,
    GetCommitMessagesRequest,
    GetTreeEntriesRequest,
    LastCommitForPathRequest,
    ListCommitsByOidRequest,
    ListCommitsByRefNameRequest,
    ListCommitsRequest,
    ListFilesRequest,
    ListLastCommitsForTreeRequest,
    RawBlameRequest,
    TreeEntry,
    TreeEntryRequest,
    TreeEntryResponse,
)
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub

from hgitaly.stub.shared_pb2 import (
    CommitStatInfo,
    GlobalOptions,
    PaginationParameter,
)

from .fixture import ServiceFixture


class CommitFixture(ServiceFixture):

    stub_cls = CommitServiceStub

    def is_ancestor(self, hex1, hex2):
        return self.stub.CommitIsAncestor(
            CommitIsAncestorRequest(repository=self.grpc_repo,
                                    ancestor_id=hex1,
                                    child_id=hex2,
                                    )
        ).value

    def find_commit(self, revision):
        return self.rpc('FindCommit',
                        FindCommitRequest(repository=self.grpc_repo,
                                          revision=revision))

    def find_commits(self, limit=10, **opts):
        request = FindCommitsRequest(repository=self.grpc_repo,
                                     limit=limit, **opts)
        resp = self.stub.FindCommits(request)
        return [commit for chunk in resp for commit in chunk.commits]

    def find_commits_ids(self, **kwargs):
        return [commit.id.encode('ascii')
                for commit in self.find_commits(**kwargs)]

    def find_commits_stats(self, **kwargs):
        return [commit.short_stats
                for commit in self.find_commits(include_shortstat=True,
                                                **kwargs)
                ]

    def count_commits(self, revision, **kwargs):
        kwargs.setdefault('repository', self.grpc_repo)
        if revision is not None:
            kwargs['revision'] = revision
        return self.rpc('CountCommits', CountCommitsRequest(**kwargs)).count

    def get_tree_entries(self, path, revision,
                         limit=10, page_token='',
                         sort=GetTreeEntriesRequest.SortBy.DEFAULT,
                         skip_flat_paths=False,
                         recursive=False):
        pagination_params = PaginationParameter(limit=limit,
                                                page_token=page_token)
        request = GetTreeEntriesRequest(repository=self.grpc_repo,
                                        revision=revision,
                                        skip_flat_paths=skip_flat_paths,
                                        path=path,
                                        pagination_params=pagination_params,
                                        sort=sort,
                                        recursive=recursive)
        entries = []
        cursor = None  # loop below can be empty
        for resp in self.stub.GetTreeEntries(request):
            entries.extend(resp.entries)
            # same implementation as on the Rails side, less robust than
            # initial version in order to reproduce hgitaly#91
            if resp.HasField('pagination_cursor'):
                cursor = resp.pagination_cursor.next_cursor

        return entries, cursor

    def list_commits(self, *revisions, limit=None, page_token=None, **kw):
        if limit is None and page_token is None:
            pagination = None
        else:
            pagination = PaginationParameter(limit=limit,
                                             page_token=page_token)

        request = ListCommitsRequest(repository=self.grpc_repo,
                                     revisions=revisions,
                                     pagination_params=pagination,
                                     **kw,
                                     )

        resp = self.stub.ListCommits(request)
        return [c.id.encode('ascii') for chunk in resp for c in chunk.commits]

    def commit_languages(self, **kw):
        request = CommitLanguagesRequest(repository=self.grpc_repo, **kw)
        return list(self.stub.CommitLanguages(request).languages)

    def last_commit_for_path(self, revision, path, **kw):
        """Call the method, returning directly commit.id, as bytes"""
        grpc_repo = kw.pop('grpc_repo', self.grpc_repo)
        request = LastCommitForPathRequest(repository=grpc_repo,
                                           revision=revision,
                                           path=path,
                                           **kw)
        return pycompat.sysbytes(
            self.stub.LastCommitForPath(request).commit.id
        )


@pytest.fixture
def commit_fixture_empty_repo(grpc_channel, server_repos_root):
    with CommitFixture(grpc_channel,
                       server_repos_root,
                       ) as fixture:
        yield fixture


@pytest.fixture
def commit_fixture_tree_shaped_repo(grpc_channel, server_repos_root):
    with CommitFixture(grpc_channel,
                       server_repos_root,
                       repo_factory=make_tree_shaped_repo,
                       ) as fixture:
        yield fixture


def test_is_ancestor(commit_fixture_tree_shaped_repo):
    fixture = commit_fixture_tree_shaped_repo
    changesets = fixture.changesets

    def is_ancestor(key1, key2):
        return fixture.is_ancestor(changesets[key1].hex(),
                                   changesets[key2].hex())

    assert is_ancestor('base', 'top1')
    assert not is_ancestor('other_base', 'default')
    assert is_ancestor('default', 'default')
    assert is_ancestor('other_base', 'wild2')

    base_hex = changesets['base'].hex()
    # id in message *has* logically to be 40 chars
    # technically, on current Mercurial if short_id is str, repo[short_id]
    # does not work but repo[full_id] does.
    unknown_hex = '1234dead' * 5

    assert fixture.is_ancestor(base_hex, unknown_hex) is False
    assert fixture.is_ancestor(unknown_hex, base_hex) is False


def test_is_ancestor_obsolete(commit_fixture_tree_shaped_repo):
    fixture = commit_fixture_tree_shaped_repo
    wrapper = fixture.repo_wrapper

    # still works if one of the changesets becomes obsolete
    ctx1 = wrapper.commit_file('foo')
    ctx2 = wrapper.commit_file('foo2')

    wrapper.command('amend', message=b'amended')
    assert fixture.is_ancestor(ctx1.hex(), ctx2.hex())


# TODO test with tags. In particular they have precedence in Mercurial
# over branches
def test_find_commit(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    fixture.correlation_id = "1111-correlation"
    wrapper = fixture.repo_wrapper
    find_commit = fixture.find_commit

    now = time.time()
    ctx = wrapper.commit_file('foo',
                              utc_timestamp=now,
                              message="A single line",
                              user="HGitaly Test <hgitaly@heptapod.test>")
    wrapper.command('tag', b'testtag', rev=ctx.hex())
    ctx2 = wrapper.commit_file('foo',
                               tz_timestamp=(123456, -7200),  # UTC+2
                               parent=ctx,
                               message="Foo deux\n\nA very interesting bar")

    commit = find_commit(ctx.hex()).commit
    assert commit is not None
    assert commit.id == ctx.hex().decode()
    assert commit.parent_ids == []
    assert commit.author.name == b"HGitaly Test"
    assert commit.author.email == b"hgitaly@heptapod.test"
    assert commit.author.date.seconds == int(now)
    assert commit.author.timezone == b'+0000'
    assert commit.subject == b"A single line"
    assert commit.body == commit.subject
    assert commit.body_size == 13

    commit2 = find_commit(ctx2.hex()).commit
    assert commit2 is not None
    assert commit2.subject == b'Foo deux'
    assert commit2.body == b"Foo deux\n\nA very interesting bar"
    assert commit2.body_size == 32
    assert commit2.parent_ids == [ctx.hex().decode()]
    assert commit2.author.timezone == b'+0200'

    # TODO check with two parents, it'd be nice to have a helper to create
    # merge commits very quickly
    assert find_commit(b'branch/default').commit == commit2

    # default GitLab branch not being set, there is no fallback
    assert not find_commit(b'HEAD').HasField('commit')

    top_ctx = wrapper.commit_file('animals',
                                  message="in topic",
                                  topic='antelope',
                                  parent=ctx)
    commit_top = find_commit(b'topic/default/antelope').commit
    assert commit_top is not None
    assert commit_top.subject == b"in topic"

    # a tag
    assert find_commit(b'testtag').commit == commit

    # with revision given in full "ref" form
    assert find_commit(
        b'refs/heads/topic/default/antelope').commit == commit_top
    assert find_commit(b'refs/tags/testtag').commit == commit

    # with explicitely set default GitLab branch:
    set_default_gitlab_branch(wrapper.repo, b'topic/default/antelope')
    assert find_commit(b'HEAD').commit == commit_top

    # with obsolete changeset given by full node id
    wrapper.command('amend', message=b'amended')
    assert find_commit(top_ctx.hex()).commit == commit_top

    # with unknown changeset
    assert not find_commit(b'unknown').HasField('commit')

    # 'tip' is not a valid tag
    # TODO avoid resolving it when given in implicit form (just 'tip')
    # but that requires not fallbacking to scmutil.revsingle and is not
    # suited for the current work on stable branch.
    assert not find_commit(b'refs/tags/tip').HasField('commit')
    assert not find_commit(b'refs/tags/unknown').HasField('commit')

    # special cases of invalid commit ids (prefix or not)
    assert not find_commit(b'f' * 39).HasField('commit')
    assert not find_commit(b'f' * 40).HasField('commit')

    # non existing valid full node
    assert not find_commit(b'dead' * 10).HasField('commit')

    # null revision
    assert not find_commit(b'0' * 40).HasField('commit')

    # no revision
    with pytest.raises(grpc.RpcError) as exc_info:
        find_commit(revision=b'')
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT


def test_find_commits(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    wrapper = fixture.repo_wrapper
    find_commits_ids = fixture.find_commits_ids
    # set default branch
    set_default_gitlab_branch(wrapper.repo, b'branch/default')
    # prepare repo as:
    #
    #   @    4 (branch/default) merge with stable
    #   |\
    #   | o  3 creates 'animal' (branch/stable)
    #   | |
    #   o |  2 rename 'foo' to 'zoo' (user: testuser2)
    #   |/
    #   | 1 changes 'foo'
    #   |/
    #   o  0  creates 'foo'
    #

    ctx0 = wrapper.commit_file('foo')
    ctx1 = wrapper.commit_file('foo')
    wrapper.update(0)
    wrapper.command(b'mv', wrapper.repo.root + b'/foo',
                    wrapper.repo.root + b'/zoo')
    ctx2 = wrapper.commit([b'foo', b'zoo'], message=b"rename foo to zoo")
    # commits with different date/time, to test with date filter later
    timestamp = int(time.time())
    ctx3 = wrapper.commit_file('animals', branch='stable', parent=ctx0,
                               utc_timestamp=timestamp+10, user='testuser2')
    wrapper.update(2)
    ctx4 = wrapper.merge_commit(ctx3, message=b'merge with stable',
                                utc_timestamp=timestamp+20)
    sha0, sha1, sha2, sha3, sha4 = (ctx0.hex(), ctx1.hex(), ctx2.hex(),
                                    ctx3.hex(), ctx4.hex())
    # symbolic revision
    assert find_commits_ids(revision=b'branch/stable') == [sha3, sha0]
    # with follow; test that it follows renames
    assert find_commits_ids(paths=[b'zoo'], follow=True) == [sha2, sha0]
    # with follow and Git Range Notation, it does not **currently** follow
    # (see hgitaly#117)
    assert find_commits_ids(revision=b'2...3', paths=[b'zoo'],
                            follow=True) == [sha2]
    # with revision
    assert find_commits_ids(revision=b'4...1') == [sha4, sha3, sha2, sha1]
    assert find_commits_ids(revision=b'0..4') == [sha4, sha3, sha2]
    # use default branch if any of the side is omitted
    assert find_commits_ids(revision=b'...1') == [sha4, sha3, sha2, sha1]
    assert find_commits_ids(revision=b'1...') == [sha4, sha3, sha2, sha1]
    assert find_commits_ids(revision=b'0..') == [sha4, sha3, sha2]
    assert find_commits_ids(revision=b'..0') == []
    # when revision does not exists (including special Node IDs)
    for rev in (b'does_not_exists',
                b'1234deadbeaf',
                b'ffffffff' * 5,
                b'ffffffff'):
        with pytest.raises(grpc.RpcError) as exc_info:
            find_commits_ids(revision=rev)
        assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND
    # with all, return all the commits
    assert find_commits_ids(all=True) == [sha0, sha1, sha2, sha3, sha4]
    # with offset
    assert find_commits_ids(all=True, offset=2) == [sha2, sha3, sha4]
    # with skip_merges
    assert find_commits_ids(skip_merges=True) == [sha3, sha2, sha0]
    # with short stats
    assert fixture.find_commits_stats(revision=sha1) == [
        CommitStatInfo(additions=2, deletions=2, changed_files=1),
        CommitStatInfo(additions=3, changed_files=1),
    ]
    assert not fixture.find_commits(revision=sha4,
                                    limit=1,
                                    include_shortstat=True
                                    )[0].HasField('short_stats')

    # with date specs
    date, date2 = Timestamp(), Timestamp()
    date.FromSeconds(timestamp+10)
    assert find_commits_ids(before=date) == [sha3, sha2, sha0]
    assert find_commits_ids(after=date) == [sha4, sha3]
    # gracinet: given that before is < and after is > I didn't expect
    # after=before to return anything, but `hg help dates` says that
    # the `to` range specifier is inclusive.
    assert find_commits_ids(after=date, before=date) == [sha3]
    date2.FromSeconds(timestamp+100)
    assert find_commits_ids(after=date, before=date2) == [sha4, sha3]

    # with pats, return commits containing changes to pats
    assert find_commits_ids(paths=[b'foo', b'animals']) == [sha3, sha0]
    # when no revision passed; return revs("reverse(::default_tip)")
    assert find_commits_ids() == [sha4, sha3, sha2, sha0]
    # with limit
    assert find_commits_ids(limit=0) == []
    assert find_commits_ids(limit=2) == [sha4, sha3]
    # with author
    assert find_commits_ids(all=True, author=b"testuser2") == [sha3]
    # with order
    ctx5 = wrapper.commit_file('foo', parent=ctx1)
    sha5 = ctx5.hex()
    assert find_commits_ids(all=True, order=FindCommitsRequest.Order.TOPO) == [
        sha0, sha2, sha3, sha4, sha1, sha5]
    # with include_referenced_by
    assert fixture.find_commits(  # without state files empty instead of crash
        revision=sha3,
        include_referenced_by=[b'refs']
    )[0].referenced_by == []
    write_gitlab_branches(wrapper.repo, {b'branch/stable': sha3})
    assert fixture.find_commits(
        revision=sha3,
        include_referenced_by=[b'refs']
    )[0].referenced_by == [b'refs/heads/branch/stable']
    assert fixture.find_commits(
        revision=sha3,
        include_referenced_by=[b'refs/tags']
    )[0].referenced_by == []
    write_gitlab_tags(wrapper.repo, {b'release': sha3})
    assert fixture.find_commits(
        revision=sha3,
        include_referenced_by=[b'refs/tags']
    )[0].referenced_by == [b'refs/tags/release']
    write_special_refs(wrapper.repo, {b'pipelines/12': sha3})
    assert fixture.find_commits(
        revision=sha3,
        include_referenced_by=[b'refs/pipelines/']
    )[0].referenced_by == [b'refs/pipelines/12']
    set_default_gitlab_branch(wrapper.repo, b'branch/stable')
    assert fixture.find_commits(
        revision=sha3,
        include_referenced_by=[b'HEAD', b'refs/heads']
    )[0].referenced_by == [b'refs/heads/branch/stable', b'HEAD']

    # tests with obsolete changesets
    wrapper.amend_file('foo')
    assert find_commits_ids(revision=sha5) == [sha5, sha1, sha0]


def test_commits_by_message(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    set_default_gitlab_branch(wrapper.repo, b'branch/default')
    # repo structure:
    #
    #   o 2 add animal (branch/stable)
    #   |
    #   | 1 add bar
    #   |/
    #   o 0 add foo
    #

    def do_rpc(**opts):
        request = CommitsByMessageRequest(repository=grpc_repo, **opts)
        resp = grpc_stub.CommitsByMessage(request)
        return [pycompat.sysbytes(commit.id)
                for chunk in resp for commit in chunk.commits]
    ctx0 = wrapper.commit_file('foo', message=b'add foo')
    ctx1 = wrapper.commit_file('bar', parent=ctx0, message=b'add bar')
    ctx2 = wrapper.commit_file('animal', branch='stable', parent=ctx0,
                               message=b'add animal')
    sha0, sha1, sha2 = ctx0.hex(), ctx1.hex(), ctx2.hex()

    # without revision; use default branch set in repo config
    assert do_rpc(query=b'add') == [sha1, sha0]
    assert do_rpc(query=b'bar') == [sha1]
    assert do_rpc(query=b'foo') == [sha0]
    # with revision
    assert do_rpc(revision=b'branch/stable', query=b'add') == [sha2, sha0]
    # when revision does not exists
    assert do_rpc(revision=b'does_not_exists', query=b'add') == []
    # with no arg
    assert do_rpc() == []
    # with offset
    assert do_rpc(query=b'add', offset=1) == [sha0]
    # with limit
    assert do_rpc(query=b'add', limit=1) == [sha1]
    # with path
    assert do_rpc(query=b'add', path=b'foo') == [sha0]

    # on obsolete changeset
    wrapper.amend_file('animal')
    assert do_rpc(revision=sha2, query=b'add') == [sha2, sha0]


def test_find_all_commits(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    # prepare repo as:
    #
    #   o  3 (branch/stable)
    #   |
    #   | o  2 user: testuser2
    #   |/
    #   | 1 changes 'foo'
    #   |/
    #   o  0  creates 'foo'
    #

    def do_rpc(**opts):
        request = FindAllCommitsRequest(repository=grpc_repo, **opts)
        resp = grpc_stub.FindAllCommits(request)
        return [pycompat.sysbytes(commit.id)
                for chunk in resp for commit in chunk.commits]
    ctx0 = wrapper.commit_file('foo')
    timestamp = int(time.time())
    # commits with different date/time, to test with date ordering
    ctx1 = wrapper.commit_file('foo', parent=ctx0, utc_timestamp=timestamp+10)
    ctx2 = wrapper.commit_file('bar', parent=ctx0, user="testuser2",
                               utc_timestamp=timestamp+20)
    ctx3 = wrapper.commit_file('animals', branch='stable', parent=ctx0,
                               utc_timestamp=timestamp+15)
    sha0, sha1, sha2, sha3 = ctx0.hex(), ctx1.hex(), ctx2.hex(), ctx3.hex()

    # with revision
    assert do_rpc(revision=b'branch/stable') == [sha3, sha0]
    assert do_rpc(revision=b'branch/default') == [sha2, sha0]
    # when revision does not exists
    assert do_rpc(revision=b'does_not_exists') == []
    # with no arg, return all the commits
    assert do_rpc() == [sha3, sha2, sha1, sha0]
    # with skip
    assert do_rpc(skip=2) == [sha1, sha0]
    # with max_count
    assert do_rpc(max_count=2) == [sha3, sha2]
    # with date order (note that ctx2 date is latest than ctx3)
    assert do_rpc(order='DATE') == [sha2, sha3, sha1, sha0]
    # with topo order
    assert do_rpc(order='TOPO') == [sha3, sha2, sha1, sha0]
    # with obsolete revision (done at the end for no repercussions)
    wrapper.prune(sha3)
    assert ctx3.obsolete()
    assert do_rpc(revision=sha3) == [sha3, sha0]


def test_list_commits_by_ref_name(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    def do_rpc(ref_names):
        request = ListCommitsByRefNameRequest(repository=grpc_repo,
                                              ref_names=ref_names)
        response = grpc_stub.ListCommitsByRefName(request)
        return {ref.ref_name: ref.commit.id.encode('ascii')
                for chunk in response for ref in chunk.commit_refs}

    # prepare repo as:
    #
    #   2 (branch/default)
    #   |
    #   1
    #   |  3 (topic/default/feature)
    #   | /
    #   0
    #
    ctx0 = wrapper.commit_file('foo')
    wrapper.commit_file('bar')
    ctx2 = wrapper.commit_file('baz')
    ctx3 = wrapper.commit_file('animals', topic='feature', parent=ctx0)

    assert do_rpc(
        [b"branch/default", b"topic/default/feature", b"not_exists"]
    ) == {
        b"branch/default": ctx2.hex(),
        b"topic/default/feature": ctx3.hex()
    }


def test_count_diverging_commits(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    def do_rpc(gl_from, gl_to, max_count=None):
        request = CountDivergingCommitsRequest(repository=grpc_repo,
                                               to=gl_to,
                                               max_count=max_count)
        setattr(request, 'from', gl_from)
        response = grpc_stub.CountDivergingCommits(request)
        return [response.left_count, response.right_count]

    # prepare repo as:
    #
    #   2 (branch/default)
    #   |
    #   1
    #   |  3 (topic/default/feature)
    #   | /
    #   0
    #
    ctx0 = wrapper.commit_file('foo')
    wrapper.commit_file('bar')
    wrapper.commit_file('baz')
    top_ctx = wrapper.commit_file('animals', topic='feature', parent=ctx0)

    assert do_rpc(b"branch/default", b"topic/default/feature") == [2, 1]
    # with commit given by full id
    assert do_rpc(b"branch/default", top_ctx.hex()) == [2, 1]
    # count 0 for the same "from" and "to"
    assert do_rpc(b"branch/default", b"branch/default") == [0, 0]
    # when one of them is invalid ref
    assert do_rpc(b"branch/default", b"does-not-exists") == [0, 0]
    assert do_rpc(b"does-not-exists", b"branch/default") == [0, 0]
    # count bounded with max_count
    for max_count in [1, 2, 3]:
        resp = do_rpc(b"branch/default", b"topic/default/feature", max_count)
        assert (resp[0] + resp[1]) == max_count
        resp = do_rpc(b"topic/default/feature", b"branch/default", max_count)
        assert (resp[0] + resp[1]) == max_count

    # with obsolete changeset
    wrapper.command('amend', message=b'amended')
    assert do_rpc(b"branch/default", top_ctx.hex()) == [2, 1]

    # with no ancestor
    wrapper.update('000000000000000000000000000')
    new_root = wrapper.commit_file('foo')
    assert do_rpc(top_ctx.hex(), new_root.hex()) == [2, 1]


def test_count_commits(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    do_rpc = fixture.count_commits

    wrapper = fixture.repo_wrapper

    # prepare repo as:
    #
    #   2 (branch/default)
    #   |
    #   1
    #   |  3 (topic/default/feature)
    #   | /
    #   0
    #
    ctx0 = wrapper.commit_file('foo')
    ctx1 = wrapper.commit_file('bar', parent=ctx0)
    wrapper.commit_file('baz', parent=ctx1)
    topic_hex = wrapper.commit_file('animals',
                                    topic='feature',
                                    parent=ctx0).hex()

    # simplest calls
    assert do_rpc(b'branch/default') == 3
    assert do_rpc(b'topic/default/feature') == 2
    assert do_rpc(topic_hex) == 2

    # when `all` is passed, return total commits in repo
    assert do_rpc(None, all=True) == 4
    assert do_rpc(None, max_count=2, all=True) == 2

    # range notation
    assert do_rpc(b'branch/default..topic/default/feature') == 1
    assert do_rpc(b'topic/default/feature..branch/default') == 2

    # cases of revision not found
    assert all(do_rpc(revision) == 0
               for revision in (b'branch/default..unknown',
                                b'unknown..topic/default/feature',
                                b'unknown'))

    # obsolescence
    wrapper.command('amend', message=b'amended')
    assert do_rpc(topic_hex) == 2
    assert do_rpc(b'branch/default..' + topic_hex) == 1
    assert do_rpc(topic_hex + b'..branch/default') == 2

    # error cases
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(None)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT


def test_last_commit_for_path(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    wrapper = fixture.repo_wrapper

    # prepare repo as:
    #
    #   1 (branch/default) changes 'foo'
    #   |
    #   |  2 (branch/other) creates 'sub/bar'
    #   | /
    #   0 creates 'foo'
    #
    ctx = wrapper.commit_file('foo')
    ctx1 = wrapper.commit_file('foo')
    # TODO should be done by testhelpers!
    (wrapper.path / 'sub').mkdir()
    ctx2 = wrapper.commit_file('sub/bar', branch='other', parent=ctx)

    lcfp = fixture.last_commit_for_path
    # case with path resolving to the repo root
    assert lcfp(revision=b'branch/other', path=b'/') == ctx2.hex()
    assert lcfp(revision=b'branch/other', path=b'') == ctx2.hex()

    assert lcfp(revision=b'branch/default', path=b'foo') == ctx1.hex()
    assert lcfp(revision=b'branch/other', path=b'foo') == ctx.hex()
    assert lcfp(revision=ctx2.hex(), path=b'foo') == ctx.hex()
    assert not lcfp(revision=b'branch/default', path=b'sub/bar')
    assert lcfp(revision=b'branch/other', path=b'sub/bar') == ctx2.hex()
    # recursive directory matching (see Rails tests for validation
    # that we must match on directories)
    assert lcfp(revision=b'branch/other', path=b'sub') == ctx2.hex()
    # with trailing slash already provided
    assert lcfp(revision=b'branch/other', path=b'sub/') == ctx2.hex()
    # pathspecs with wildcards
    assert lcfp(revision=b'branch/other', path=b's*/bar') == ctx2.hex()
    assert lcfp(revision=b'branch/other', path=b's*r') == ctx2.hex()
    # literal pathspecs
    literal = GlobalOptions(literal_pathspecs=True)
    assert not lcfp(b'branch/other', b's*/bar', global_options=literal)
    assert lcfp(b'branch/other', b'sub', global_options=literal) == ctx2.hex()
    assert lcfp(b'branch/other', b'sub/', global_options=literal) == ctx2.hex()

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx2.obsolete()
    assert lcfp(revision=ctx2.hex(), path=b'foo') == ctx.hex()

    # with a removal
    (wrapper.path / 'sub/bar').unlink()
    ctx_rm = wrapper.commit(rel_paths=['sub/bar'], message="removed",
                            add_remove=True)
    wrapper.commit_file('foo2')
    assert lcfp(revision=b'branch/other', path=b'sub') == ctx_rm.hex()

    # error response for unresolvable revision
    with pytest.raises(grpc.RpcError) as exc_info:
        lcfp(revision=b'unknown', path=b'foo') == b''
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL


def test_list_commits_by_oid(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    def do_rpc(oids):
        request = ListCommitsByOidRequest(repository=grpc_repo,
                                          oid=oids)
        resp = grpc_stub.ListCommitsByOid(request)
        return [pycompat.sysbytes(commit.id)
                for chunk in resp for commit in chunk.commits]

    ctx = wrapper.commit_file('foo')
    ctx2 = wrapper.commit_file('foo')

    # with 39, chances of collision are very thin and it still demonstrates
    # that prefix lookup works
    short = ctx.hex()[:39]
    short2 = ctx2.hex()[:39]

    assert do_rpc([short]) == [ctx.hex()]
    assert do_rpc([short, short2]) == [ctx.hex(), ctx2.hex()]

    assert do_rpc([]) == []

    unknown_sha = "6caf" * 9  # prefix of length 36, very unlikey to match
    assert do_rpc([unknown_sha]) == []
    assert do_rpc([unknown_sha, short]) == [ctx.hex()]

    # heptapod#717, both code paths
    assert do_rpc(["ffffffff"]) == []
    assert do_rpc([unknown_sha, "ffffffff", short]) == [ctx.hex()]

    # heptapod#640: obsolete changeset given by its full hash
    wrapper.command('amend', message=b'amended')
    assert do_rpc([ctx2.hex()]) == [ctx2.hex()]


def test_get_commit_messages(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    # prepare repo as:
    #
    #   1 (branch/default) changes 'foo'
    #   |
    #   0 creates 'foo'
    #
    ctx0 = wrapper.commit_file('foo', message=b'added foo')
    ctx1 = wrapper.commit_file('foo', message=b'changes foo')
    sha0, sha1 = ctx0.hex(), ctx1.hex()

    def do_rpc(commit_ids):
        request = GetCommitMessagesRequest(repository=grpc_repo,
                                           commit_ids=commit_ids)
        response = grpc_stub.GetCommitMessages(request)
        results = {}
        for resp in response:
            commit_id = resp.commit_id
            if commit_id:
                results[commit_id] = resp.message
        return list(results.values())
    assert do_rpc([sha1, sha0]) == [b"changes foo", b"added foo"]
    assert do_rpc([sha1]) == [b"changes foo"]

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc([sha1, '12ca34de' * 5, sha0])
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'object not found' in exc_info.value.details()

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx1.obsolete()
    assert do_rpc([sha1]) == [b"changes foo"]


def test_commit_stats(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    # repo structure:
    #
    #   2 Removed two lines
    #   |
    #   1 added one line, removed one
    #   |
    #   0 added foo
    #
    sha0 = wrapper.commit_file(
        'foo',
        content="First line\nSecond line\nThird line\n",
        message=b'added foo'
    ).hex()

    sha1 = wrapper.commit_file(
        'foo',
        content="Second line\nThird line\nFourth line\n",
        message=b'added one line, removed one'
    ).hex()

    ctx2 = wrapper.commit_file(
        'foo',
        content="Fourth line\n",
        message=b'Removed two lines'
    )
    sha2 = ctx2.hex()

    def do_rpc(rev):
        request = CommitStatsRequest(repository=grpc_repo,
                                     revision=rev)
        resp = grpc_stub.CommitStats(request)
        return resp.additions, resp.deletions

    assert do_rpc(sha0) == (3, 0)
    assert do_rpc(sha1) == (1, 1)
    assert do_rpc(sha2) == (0, 2)

    # with unknown revision
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(b"23fire32" * 5)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'object not found' in exc_info.value.details()

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx2.obsolete()
    assert do_rpc(sha2) == (0, 2)


def test_tree_entry(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    ctx0 = wrapper.commit_file('foo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    (sub / 'ba2').chmod(0o755)
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                          message="zebar", add_remove=True)

    def do_rpc(path, limit=0, max_size=0, revision=ctx1.hex()):
        request = TreeEntryRequest(repository=grpc_repo,
                                   revision=revision,
                                   limit=limit,
                                   max_size=max_size,
                                   path=path)
        resp = grpc_stub.TreeEntry(request)
        return [r for r in resp]

    def rpc_single_response(*args, **kwargs):
        resps = do_rpc(*args, **kwargs)
        assert len(resps) == 1
        return resps[0]

    for rev in (ctx1.hex(), b'branch/default'):
        resp = rpc_single_response(b'sub/bar', revision=rev)
        assert resp.data == b'bar content'
        assert resp.mode == 0o100644
        assert resp.type == TreeEntryResponse.ObjectType.BLOB

    resp = rpc_single_response(b'sub')
    assert resp.type == TreeEntryResponse.ObjectType.TREE
    assert resp.mode == 0o40000

    resp = rpc_single_response(b'sub/ba2', limit=3)
    assert resp.type == TreeEntryResponse.ObjectType.BLOB
    assert resp.data == b'ba2'

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(b'foo', revision=ctx0.hex(), max_size=5)
    assert exc_info.value.code() == grpc.StatusCode.FAILED_PRECONDITION
    err_details = exc_info.value.details()
    assert 'bigger than the maximum allowed size (5)' in err_details

    # revision or file not found
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(b'not/there')
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(b'foo', revision=b'branch/unknown')
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx1.obsolete()
    resp = rpc_single_response(b'sub/bar', revision=rev)
    assert resp.data == b'bar content'
    assert resp.mode == 0o100644
    assert resp.type == TreeEntryResponse.ObjectType.BLOB

    # chunking
    thirty_two = b'ba' * 16384
    for content_suffix in (b'', b'x'):
        content = thirty_two + content_suffix
        rev_hex = wrapper.commit_file(b'bar', content).hex()
        resps = do_rpc(b'bar', revision=rev_hex)

        assert len(resps) == 3 if content_suffix else 2
        # metadata are carried by the first response only
        assert resps[0].type == TreeEntryResponse.ObjectType.BLOB
        assert resps[0].mode == 0o100644
        for r in resps[1:]:
            assert not r.type
            assert not r.mode

        assert b''.join(r.data for r in resps) == content


def test_get_tree_entries(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    wrapper = fixture.repo_wrapper
    repo = wrapper.repo
    get_tree_entries = fixture.get_tree_entries

    wrapper.commit_file('foo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    (sub / 'ba2').chmod(0o755)
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                          message="zebar", add_remove=True)

    SortBy = GetTreeEntriesRequest.SortBy

    ctx1_oid = ctx1.hex().decode()
    sha1 = ctx1.hex()
    bar_default = TreeEntry(path=b'sub/bar',
                            flat_path=b'sub/bar',
                            mode=0o100644,
                            oid=blob_oid(repo, ctx1_oid, b'sub/bar'),
                            commit_oid=ctx1_oid,
                            )
    ba2_default = TreeEntry(path=b'sub/ba2',
                            flat_path=b'sub/ba2',
                            mode=0o100755,
                            oid=blob_oid(repo, ctx1_oid, b'sub/ba2'),
                            commit_oid=ctx1_oid,
                            )
    foo_default = TreeEntry(path=b'foo',
                            flat_path=b'foo',
                            mode=0o100644,
                            oid=blob_oid(repo, ctx1_oid, b'foo'),
                            commit_oid=ctx1_oid,
                            )
    sub_default = TreeEntry(path=b'sub',
                            flat_path=b'sub',
                            mode=0o40000,
                            type=TreeEntry.EntryType.TREE,
                            oid=blob_oid(repo, ctx1_oid, b'sub'),
                            commit_oid=ctx1_oid,
                            )

    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'', revision=sha1)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'empty path' in exc_info.value.details()

    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'sub/bar', revision=sha1)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'no-such-dir', revision=sha1)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'no-such-dir', revision=sha1, recursive=True)
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND

    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'.', revision=b'unknown')
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'invalid revision' in exc_info.value.details()

    assert get_tree_entries(b'sub', revision=sha1) == ([ba2_default,
                                                        bar_default], '')

    assert get_tree_entries(b'.', revision=sha1, recursive=False)[0] == [
        foo_default, sub_default]
    # Note for recursive call: that top-level paths are before deeper
    # ones is a mere coincidence.
    expected_recursive = [foo_default, sub_default, ba2_default, bar_default]
    for entry in expected_recursive:
        entry.flat_path = b''
    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True)[0] == expected_recursive

    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True, limit=0)[0] == []
    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True, limit=-1)[0] == expected_recursive

    limited, cursor = get_tree_entries(b'.', revision=sha1,
                                       recursive=True, limit=2)
    assert limited == expected_recursive[:2]
    assert cursor == expected_recursive[1].oid
    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True,
                            page_token=cursor)[0] == expected_recursive[2:]

    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True, sort=SortBy.TREES_FIRST)[0] == [
        sub_default, foo_default, ba2_default, bar_default]

    invalid_cursor = 'this is not an OID'
    with pytest.raises(grpc.RpcError) as exc_info:
        get_tree_entries(b'.', revision=sha1, page_token=invalid_cursor)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert invalid_cursor in exc_info.value.details()

    # with skip_flat_paths, even non recursive calls give empty flat paths
    assert get_tree_entries(b'.', revision=sha1,
                            skip_flat_paths=True)[0] == [foo_default,
                                                         sub_default]

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx1.obsolete()
    assert get_tree_entries(b'.', revision=sha1,
                            recursive=True)[0] == expected_recursive


def test_list_last_commits_for_tree(grpc_channel, server_repos_root):

    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    ctx0 = wrapper.commit_file('foo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                          message="zebar", add_remove=True)
    ctx2 = wrapper.commit_file('sub/bar')
    ctx3 = wrapper.commit_file('foo', parent=ctx1, branch='other')
    ctx4 = wrapper.commit_file('oof')

    sha0 = ctx0.hex().decode()
    sha1 = ctx1.hex().decode()
    sha2 = ctx2.hex().decode()
    sha3 = ctx3.hex().decode()
    sha4 = ctx4.hex().decode()

    def do_rpc(rev, path, offset=0, limit=None):
        request = ListLastCommitsForTreeRequest(
            repository=grpc_repo,
            offset=offset,
            limit=1000 if limit is None else limit,
            revision=rev,
            path=path)
        return [(cft.path_bytes, cft.commit.id)
                for chunk in grpc_stub.ListLastCommitsForTree(request)
                for cft in chunk.commits]

    assert do_rpc(sha2, b'sub/', limit=0) == []

    assert do_rpc(sha2, b'sub/') == [
        (b'sub/ba2', sha1),
        (b'sub/bar', sha2),
    ]

    # we need the trailing slash to trigger inspection of tree content
    assert do_rpc(sha2, b'sub') == [(b'sub', sha2)]
    assert do_rpc(sha2, b'sub', offset=1) == []

    # ...except for the root tree
    assert do_rpc(sha2, b'') == [
        (b'sub', sha2),
        (b'foo', sha0),
    ]

    # other spellings of the root tree
    assert do_rpc(sha2, b'.') == do_rpc(sha2, b'')
    assert do_rpc(sha2, b'/') == do_rpc(sha2, b'')
    assert do_rpc(sha2, b'./') == do_rpc(sha2, b'')

    # usage of revision instead of a sha
    assert do_rpc('branch/default', b'sub') == do_rpc(sha2, b'sub')

    assert do_rpc(sha3, b'sub/') == [
        (b'sub/ba2', sha1),
        (b'sub/bar', sha1),
    ]

    assert do_rpc(sha3, b'.') == [
        (b'sub', sha1),
        (b'foo', sha3),
    ]

    # for offset, limit tests we'll use sha4 to have more content.
    assert do_rpc(sha4, b'', limit=4) == [
        (b'sub', sha1),
        (b'foo', sha3),
        (b'oof', sha4),
    ]

    assert do_rpc(sha4, b'', limit=2) == [
        (b'sub', sha1),
        (b'foo', sha3),
    ]

    assert do_rpc(sha4, b'', offset=1, limit=2) == [
        (b'foo', sha3),
        (b'oof', sha4),
    ]

    # The null node is not really unknown, but Gitaly result is the same
    for unknown_sha in (b'be0123ef' * 5, NULL_HEX):
        with pytest.raises(grpc.RpcError) as exc_info:
            do_rpc(unknown_sha, b'')
        assert exc_info.value.code() == grpc.StatusCode.INTERNAL
        assert 'exit status 128' in exc_info.value.details()

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(sha1, b'', limit=-1)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'limit' in exc_info.value.details()

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(sha1, b'', offset=-1)
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'offset' in exc_info.value.details()

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert ctx4.obsolete()
    assert do_rpc(sha4, b'', limit=2) == [
        (b'sub', sha1),
        (b'foo', sha3),
    ]

    # case when we don't have to walk the entire changelog
    (wrapper.path / 'more').mkdir()
    sha5 = wrapper.commit_file('more/foo').hex().decode()
    assert do_rpc(sha5, b'more/') == [
        (b'more/foo', sha5),
    ]

    # case of limit smaller than number of subtrees, now that sha5 created
    # a second one
    assert do_rpc(sha5, b'', limit=1) == [
        (b'more', sha5),
    ]
    # some more to double check offset by one potential errors
    assert do_rpc(sha5, b'', offset=1, limit=1) == [
        (b'sub', sha1),
    ]
    assert do_rpc(sha5, b'', offset=1, limit=2) == [
        (b'sub', sha1),
        (b'foo', sha3),
    ]


def test_raw_blame(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    # repo structure:
    #
    #   2 updated both lines
    #   |
    #   1 added one more line
    #   |
    #   0 added first line to foo
    #
    sha0 = wrapper.commit_file(
        'foo', message=b'zefoo',
        content='Second line\n'
                'Third line\n'
    ).hex()

    sha1 = wrapper.commit_file(
        'foo', message=b'zefoo',
        content='First line\n'
                'Second line\n'
                'Third line\n'
                'Forth line\n'
    ).hex()

    def do_rpc(rev, path, **kw):
        request = RawBlameRequest(repository=grpc_repo,
                                  revision=rev,
                                  path=path,
                                  **kw)
        response = grpc_stub.RawBlame(request)
        data = [resp.data for resp in response]
        return b''.join(data)

    assert do_rpc(sha0, b'foo') == (
        b'%s 1 1\n'
        b'\tSecond line\n'
        b'%s 2 2\n'
        b'\tThird line\n' % (sha0, sha0)
    )

    expected_rev1 = (
        b'%s 1 1\n'
        b'\tFirst line\n'
        b'%s 1 2\n'
        b'\tSecond line\n'
        b'%s 2 3\n'
        b'\tThird line\n'
        b'%s 4 4\n'
        b'\tForth line\n' % (sha1, sha0, sha0, sha1)
    )
    assert do_rpc(sha1, b'foo') == expected_rev1

    expected_rev1_range = (
        b'%s 1 1\n'
        b'\tSecond line\n'
        b'%s 2 2\n'
        b'\tThird line\n' % (sha0, sha0)
    )
    assert do_rpc(sha1, b'foo', range=b'2,3') == expected_rev1_range

    # with unknown revision
    assert do_rpc(b"23fire32" * 5, b'foo') == b''

    # with empty path
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(sha0, b'')
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'empty Path' in exc_info.value.details()

    # with unknown path
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(sha0, b'unknown')
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND
    assert 'not found in revision' in exc_info.value.details()

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert do_rpc(sha1, b'foo') == expected_rev1

    # bad range
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(sha1, b'foo', range=b'100,200')


def test_list_files(grpc_channel, server_repos_root):
    grpc_stub = CommitServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    sha0 = wrapper.commit_file('foo').hex()
    sha1 = wrapper.commit_file('bar').hex()
    sha2 = wrapper.commit_file('zoo').hex()

    def do_rpc(rev):
        request = ListFilesRequest(repository=grpc_repo,
                                   revision=rev)
        response = grpc_stub.ListFiles(request)
        final = []
        for resp in response:
            final.extend(resp.paths)
        return final

    assert do_rpc(sha0) == [b'foo']
    assert do_rpc(sha1) == [b'bar', b'foo']
    assert do_rpc(sha2) == [b'bar', b'foo', b'zoo']

    # with unknown revision
    assert do_rpc(b"23fire32" * 5) == []

    # special cases of invalid commit ids (prefix or full)
    assert do_rpc(b"f" * 39) == []
    assert do_rpc(b"f" * 40) == []

    # with obsolescence
    wrapper.command('amend', message=b'amended')
    assert do_rpc(sha2) == [b'bar', b'foo', b'zoo']


def test_list_commits(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    wrapper = fixture.repo_wrapper
    list_commits = fixture.list_commits

    # see graph in Gitaly Comparison test
    ctx0 = wrapper.commit_file('foo', message="root changeset")
    sha0 = ctx0.hex()
    ctx1 = wrapper.commit_file('foo', topic='sampletop')
    wrapper.update(ctx0.rev())
    wrapper.command(b'mv', wrapper.repo.root + b'/foo',
                    wrapper.repo.root + b'/zoo')
    ts = int(time.time())
    sha2 = wrapper.commit([b'foo', b'zoo'],
                          message=b"rename foo to zoo",
                          utc_timestamp=ts - 10,).hex()
    ctx3 = wrapper.commit_file('animals', branch='stable', parent=ctx0,
                               utc_timestamp=ts+10,
                               user='testuser <testuser@heptapod.test')
    sha3 = ctx3.hex()
    wrapper.update(2)
    ctx4 = wrapper.merge_commit(ctx3, message=b'merge with stable')
    sha4 = ctx4.hex()

    def caret(ctx):
        return b'^' + ctx.hex()

    assert list_commits(sha4, caret(ctx1)) == [sha4, sha3, sha2]
    assert list_commits(sha4, caret(ctx1), reverse=True) == [sha2, sha3, sha4]

    # limit and page token
    assert list_commits(sha4, limit=2) == [sha4, sha3]
    assert list_commits(sha4, limit=10, page_token=sha2) == [ctx0.hex()]

    # orderings
    Order = ListCommitsRequest.Order
    assert list_commits(sha4, order=Order.TOPO) == [sha4, sha3, sha2, sha0]
    assert list_commits(sha4, order=Order.DATE) == [sha2, sha0, sha4, sha3]

    # no result
    assert list_commits(sha4, caret(ctx4)) == []

    # message patterns
    assert list_commits(sha4, commit_message_patterns=[b'foo']) == [sha2]
    assert set(
        list_commits(sha4, commit_message_patterns=[b'foo', b'root'])
    ) == {sha0, sha2}
    assert list_commits(sha4,
                        commit_message_patterns=[b'FOO'],
                        ignore_case=True) == [sha2]

    # unknown revision, including special cases:
    for unknown in ('branch/unknown',
                    'f' * 39,  # wdir node id, prefix
                    'f' * 40,  # wdir full node id
                    ):
        with pytest.raises(grpc.RpcError) as exc_info:
            list_commits(as_bytes(unknown))

        assert exc_info.value.code() == grpc.StatusCode.INTERNAL
        assert unknown in exc_info.value.details()

    # invalid arguments
    with pytest.raises(grpc.RpcError) as exc_info:
        list_commits()
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'revision' in exc_info.value.details()

    with pytest.raises(grpc.RpcError) as exc_info:
        list_commits(sha0, commit_message_patterns=[b'+'])
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'invalid regexp' in exc_info.value.details()

    # with obsolete changeset
    wrapper.command('amend', message=b'amended')
    assert ctx4.obsolete()
    assert list_commits(sha4, limit=2) == [sha4, sha3]


def test_commit_languages(commit_fixture_empty_repo):
    fixture = commit_fixture_empty_repo
    wrapper = fixture.repo_wrapper

    # a Python file with valid content, including a blank line
    wrapper.commit_file('foo.py',
                        content="import sys\n\nprint(sys.version)\n",
                        message="Some Python",
                        )
    # a Ruby file with valid content, with a comment
    wrapper.commit_file('bar.rb',
                        content=("# frozen_string_literal: true\n"
                                 "\n"
                                 "module Bar\n"
                                 "def f(x)\n"
                                 "  x+1\n"
                                 ),
                        message="Some Ruby",
                        )

    Language = CommitLanguagesResponse.Language
    assert fixture.commit_languages() == [
        Language(name='Ruby',
                 bytes=57,
                 share=64.7727279663086,
                 color="#701516"),
        Language(name='Python',
                 bytes=31,
                 share=35.227272033691406,
                 color="#3572A5"),
    ]

    # color defaulting: COBOL does not have a color in github-linguist's
    # `languages.json`
    wrapper.commit_file(
        'truc.cob',
        content="Not really pretending to be COBOL, "
        "extension should be enough")
    assert fixture.commit_languages()[0] == Language(name='COBOL',
                                                     bytes=61,
                                                     share=40.939598083496094,
                                                     color="#d29900")

    # test with a subrepo, see heptapod#1287
    nested_path = wrapper.path / 'nested'
    nested = RepoWrapper.init(nested_path)
    nested.commit_file("bar", content="in nested")
    (wrapper.path / '.hgsub').write_bytes(b"nested = nested\n")
    wrapper.command(b'add', subrepos=True)
    wrapper.command(b'commit', subrepos=True, message=b"invalid")

    assert fixture.commit_languages() == []



# Copyright 2020-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# flake8: noqa F401
from .address import (
    InvalidUrl,
    UnsupportedUrlScheme,
)
from .mono import (
    BindError,
)
from .prefork import (
    run_forever,
)



# Copyright 2021 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from io import BytesIO
import logging

import attr
from grpc import StatusCode
from mercurial import (
    node,
    patch as patchmod,
    pycompat,
    diffutil,
    cmdutil,
    scmutil,
)

from ..diff import (
    Chunk,
    changed_paths,
    chunk_stats,
    diff_opts,
    git_patch_id,
)
from ..git import (
    EMPTY_TREE_OID,
)
from ..logging import LoggerAdapter
from ..oid import (
    split_chgsid_path,
)
from ..revision import gitlab_revision_changeset
from ..servicer import HGitalyServicer
from ..stream import (
    aggregate_flush_batch,
    concat_resplit,
    iter_boolean_lookahead,
    split_batches,
    WRITE_BUFFER_SIZE,
)
from ..stub.diff_pb2 import (
    CommitDiffRequest,
    CommitDiffResponse,
    CommitDelta as MsgCommitDelta,
    CommitDeltaRequest,
    CommitDeltaResponse,
    FindChangedPathsRequest,
    FindChangedPathsResponse,
    GetPatchIDRequest,
    GetPatchIDResponse,
    RawDiffRequest,
    RawDiffResponse,
    RawPatchRequest,
    RawPatchResponse,
    DiffStatsRequest,
    DiffStatsResponse,
)
from ..stub.diff_pb2_grpc import DiffServiceServicer
from ..util import chunked


base_logger = logging.getLogger(__name__)
# Copied from mercurial/patch.py
# TODO even though, this is bytes and not octal, would be better
# to use the constants in `hgitaly.git`
gitmode = {b'l': b'120000', b'x': b'100755', b'': b'100644'}
nullid = node.nullid
nullhex = node.nullhex
nullrev = node.nullrev
hex = node.hex

DIFF_MSG_SIZE_THRESHOLD = 5 * 1024
"""DIFF_MSG_SIZE_THRESHOLD controls the max size of response.raw_patch_data
As per Gitaly, this threshold is used only by CommitDiff and CommitDelta
"""

MAX_NUM_STAT_BATCH_SIZE = 1000
"""MAX_NUM_STAT_BATCH_SIZE limits the number of stats in a batch
"""
MAX_FILES_UPPERBOUND = 5000
"""MAX_FILES_UPPERBOUND controls how much max_files limit can reach"""
MAX_LINES_UPPERBOUND = 250000
"""MAX_LINES_UPPERBOUND controls how much max_lines limit can reach"""
MAX_BYTES_UPPERBOUND = 5000 * 5120  # 24MB
"""MAX_BYTES_UPPERBOUND controls how much max_bytes limit can reach"""
SAFE_MAX_FILES_UPPERBOUND = 500
"""SAFE_MAX_FILES_UPPERBOUND controls how much safe_max_files limit can reach
"""
SAFE_MAX_LINES_UPPERBOUND = 25000
"""SAFE_MAX_LINES_UPPERBOUND controls how much safe_max_lines limit can reach
"""
SAFE_MAX_BYTES_UPPERBOUND = 500 * 5120  # 2.4MB
"""SAFE_MAX_BYTES_UPPERBOUND controls how much safe_max_bytes limit can reach
"""
MAX_PATCH_BYTES_UPPERBOUND = 512000  # 500KB
"""MAX_PATCH_BYTES_UPPERBOUND controls how much max_patch_bytes limit can reach
"""


@attr.s
class Limits:
    # If true, max{_files,_lines,_bytes} will cause parsing to stop if any of
    # these limits is reached
    enforce_limits = attr.ib(False)
    # Number of maximum files to parse. The file parsed after this limit is
    # reached is marked as the overflow.
    max_files = attr.ib(0)
    # Number of diff lines to parse (including lines preceded with --- or +++).
    # The file in which limit is reached is discarded and marked as overflow.
    max_lines = attr.ib(0)
    # Number of bytes to parse (including lines preceded with --- or +++).
    # The file in which limit is reached is discarded and marked as overflow.
    max_bytes = attr.ib(0)
    # Number of bytes a single patch can have. Patches surpassing this limit
    # are pruned / nullified.
    max_patch_bytes = attr.ib(0)
    # If true, safe_max{_files,_lines,_bytes} will cause diffs to collapse
    # (i.e. patches are emptied) after any of these limits reached
    collapse_diffs = attr.ib(False)
    # Number of files to parse, after which all subsequent files are collapsed.
    safe_max_files = attr.ib(0)
    # Number of lines to parse (including lines preceded with --- or +++),
    # after which all subsequent files are collapsed.
    safe_max_lines = attr.ib(0)
    # Number of bytes to parse (including lines preceded with --- or +++),
    # after which all subsequent files are collapsed.
    safe_max_bytes = attr.ib(0)


# _enforce_limits_upperbound ensures every limit value is within its
# corresponding upperbound
def _enforce_limits_upperbound(lim):
    lim.max_files = min(lim.max_files, MAX_FILES_UPPERBOUND)
    lim.max_lines = min(lim.max_lines, MAX_LINES_UPPERBOUND)
    lim.max_bytes = min(lim.max_bytes, MAX_BYTES_UPPERBOUND)
    lim.safe_max_files = min(lim.safe_max_files, SAFE_MAX_FILES_UPPERBOUND)
    lim.safe_max_lines = min(lim.safe_max_lines, SAFE_MAX_LINES_UPPERBOUND)
    lim.safe_max_bytes = min(lim.safe_max_bytes, SAFE_MAX_BYTES_UPPERBOUND)
    lim.max_patch_bytes = min(lim.max_patch_bytes, MAX_PATCH_BYTES_UPPERBOUND)


@attr.s
class CurrDiff:
    bytes_count = attr.ib(0)
    collapsed = attr.ib(False)
    line_count = attr.ib(0)
    patch = attr.ib(BytesIO())
    too_large = attr.ib(False)


@attr.s
class Parser:
    """Parser holds necessary state for parsing a diff stream"""
    limits = attr.ib()
    curr_diff = attr.ib()
    files_processed = attr.ib(0)
    lines_processed = attr.ib(0)
    bytes_processed = attr.ib(0)
    finished = attr.ib(False)
    overflow_marker = attr.ib(False)

    # In principle, "exceeded" should be > and not >=, but that's what Gitaly's
    # Golang implementation does. Looks like only `safe_max_lines_exceeded`
    # truly respecting this, for now.
    def max_files_exceeded(self):
        return self.files_processed >= self.limits.max_files

    def safe_max_files_exceeded(self):
        return self.files_processed >= self.limits.safe_max_files

    def max_bytes_exceeded(self):
        total_bytes = self.bytes_processed + self.curr_diff.bytes_count
        return total_bytes >= self.limits.max_bytes

    def safe_max_bytes_exceeded(self):
        total_bytes = self.bytes_processed + self.curr_diff.bytes_count
        return total_bytes > self.limits.safe_max_bytes

    def max_lines_exceeded(self):
        total_lines = self.lines_processed + self.curr_diff.line_count
        return total_lines >= self.limits.max_lines

    def safe_max_lines_exceeded(self):
        total_lines = self.lines_processed + self.curr_diff.line_count
        return total_lines > self.limits.safe_max_lines

    def is_over_limits(self):
        return (
            self.max_files_exceeded()
            or self.max_bytes_exceeded()
            or self.max_lines_exceeded()
        )

    def is_over_safe_limits(self):
        return (
            self.safe_max_files_exceeded()
            or self.safe_max_bytes_exceeded()
            or self.safe_max_lines_exceeded()
        )

    def parse(self, iter_file_hunks):
        # For explicitness, let's instantiate a new BytesIO obj for each file
        self.curr_diff.patch = BytesIO()
        for diffhunk in iter_file_hunks:
            self.curr_diff.patch.write(diffhunk)
            if diffhunk.startswith(b'@@'):
                diffhunk = diffhunk.split(b'\n', 1)[-1]
            lines = diffhunk.splitlines(True)
            self.curr_diff.bytes_count += len(diffhunk)
            self.curr_diff.line_count += len(lines)

        if self.limits.collapse_diffs:
            if self.is_over_safe_limits():
                self.curr_diff.patch = BytesIO()
                self.curr_diff.collapsed = True

        if self.limits.enforce_limits:
            # Apply single-file size limit
            # Note: here we are not comparing with curr_diff.bytes_count
            curr_patch = self.curr_diff.patch.getvalue()
            if len(curr_patch) >= self.limits.max_patch_bytes:
                self.curr_diff.patch = BytesIO()
                self.curr_diff.too_large = True

            if self.is_over_limits():
                self.finished = True
                self.overflow_marker = True

    def consume_curr_diff(self):
        self.files_processed += 1
        self.bytes_processed += self.curr_diff.bytes_count
        self.lines_processed += self.curr_diff.line_count
        self.curr_diff = CurrDiff()


def get_commitDiff_parser(request):
    limits = Limits()
    if request.enforce_limits:
        limits.enforce_limits = True
        limits.max_files = request.max_files
        limits.max_lines = request.max_lines
        limits.max_bytes = request.max_bytes
        limits.max_patch_bytes = request.max_patch_bytes
    if request.collapse_diffs:
        limits.collapse_diffs = True
        limits.safe_max_files = request.safe_max_files
        limits.safe_max_lines = request.safe_max_lines
        limits.safe_max_bytes = request.safe_max_bytes
    _enforce_limits_upperbound(limits)
    return Parser(limits, CurrDiff())


WhiteSpaceChanges = CommitDiffRequest.WhitespaceChanges
WHITESPACE_OPTS_MAPPING = {
    WhiteSpaceChanges.WHITESPACE_CHANGES_IGNORE: b'ignore_space_change',
    WhiteSpaceChanges.WHITESPACE_CHANGES_IGNORE_ALL: b'ignore_all_space',
}
"""Maps whitespace_changes in DiffRequest to opts for `hg diff`.

Yes, it turns out that options of `hg diff` and `git diff` have exactly
the same name.
"""


class DiffServicer(DiffServiceServicer, HGitalyServicer):
    """DiffService implementation.

    Note: There is a case where we can have differences in HGitaly's and
    Gitaly's DiffService responses. This happens with the renames when the
    similarity index is less than 50% in that case Git doesn't consider it
    a 'rename' but Hg does (as Hg has explicit tracking of copies and renames).
    """
    def CommitDiff(self, request: CommitDiffRequest,
                   context) -> CommitDiffResponse:
        parsed_request = parse_diff_request(self, request, context)
        parsed, repo, ctx_from, ctx_to = parsed_request
        if not parsed:
            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")

        m = None
        if request.paths:
            m = scmutil.matchfiles(repo, request.paths)
        parser = get_commitDiff_parser(request)
        opts = {b'git': True}
        # tell to interpret whitespace
        whitespace = bool(request.whitespace_changes)
        if whitespace:
            opts[WHITESPACE_OPTS_MAPPING[request.whitespace_changes]] = True
        # XXX: add support and tests for request.ignore_whitespace_change
        diffopts = diffutil.difffeatureopts(
            repo.ui,
            opts=opts,
            git=True,
            whitespace=whitespace,
        )
        diffchunks = ctx_to.diff(
            ctx_from,
            match=m,
            opts=diffopts
        )
        for header in patchmod.parsepatch(diffchunks):
            chunk = Chunk(header, ctx_from, ctx_to)
            from_path, to_path = chunk.from_to_file_paths
            from_id, to_id = chunk.from_to_blob_oids()
            old_mode, new_mode = chunk.from_to_file_mode()
            # For CommitDiffResponse, modes are returned in decimal form
            old_mode, new_mode = int(old_mode, 8), int(new_mode, 8)

            # generator func to yield hunks
            def in_chunks():
                for hunk in header.hunks:
                    with BytesIO() as extracted:
                        hunk.write(extracted)
                        yield extracted.getvalue()

            parser.parse(in_chunks())

            if parser.finished and parser.overflow_marker:
                yield CommitDiffResponse(
                    end_of_patch=True,
                    overflow_marker=True,
                )
                return
            else:
                curr_diff = parser.curr_diff
                response = CommitDiffResponse(
                    from_path=from_path,
                    to_path=to_path,
                    from_id=from_id,
                    to_id=to_id,
                    old_mode=old_mode,
                    new_mode=new_mode,
                    binary=header.binary(),
                    too_large=curr_diff.too_large,
                    collapsed=curr_diff.collapsed,
                )
                patch = curr_diff.patch.getvalue()
                patch_itr = split_batches(patch, DIFF_MSG_SIZE_THRESHOLD)
                for p, eop in iter_boolean_lookahead(patch_itr):
                    response.raw_patch_data = p
                    response.end_of_patch = eop
                    yield response
                    # Use a new response so we don't send other
                    # fields (from_path, ...) over and over
                    response = CommitDiffResponse()
                parser.consume_curr_diff()

    def CommitDelta(self, request: CommitDeltaRequest,
                    context) -> CommitDeltaResponse:
        parsed_request = parse_diff_request(self, request, context)
        _parsed, repo, ctx_from, ctx_to = parsed_request
        if not _parsed:
            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")
        m = None
        if request.paths:
            m = scmutil.matchfiles(repo, request.paths)
        opts = {b'git': True}
        diffopts = diffutil.difffeatureopts(repo.ui, opts=opts, git=True)
        diffchunks = ctx_to.diff(
            ctx_from,
            match=m,
            opts=diffopts
        )

        def in_deltas():
            for header in patchmod.parsepatch(diffchunks):
                chunk = Chunk(header, ctx_from, ctx_to)
                from_path, to_path = chunk.from_to_file_paths
                from_id, to_id = chunk.from_to_blob_oids()
                old_mode, new_mode = chunk.from_to_file_mode()
                # For CommitDeltaResponse, modes are returned in decimal form
                old_mode, new_mode = int(old_mode, 8), int(new_mode, 8)
                # As per Gitaly/Git behavior, if current Delta is a Rename and
                # if `from_path` is missing in `request.paths`, the other file
                # should be considered as Added instead of a Rename
                if (request.paths and from_path != to_path and
                        from_path not in request.paths):
                    # considering `to_path` as added
                    from_path = to_path
                    from_id = nullhex
                    old_mode = 0
                delta = MsgCommitDelta(
                    from_path=from_path,
                    to_path=to_path,
                    from_id=from_id,
                    to_id=to_id,
                    old_mode=old_mode,
                    new_mode=new_mode,
                )
                yield delta
        for batch in aggregate_flush_batch(in_deltas(), commit_delta_size,
                                           DIFF_MSG_SIZE_THRESHOLD):
            yield CommitDeltaResponse(deltas=batch)

    def RawDiff(self, request: RawDiffRequest,
                context) -> RawDiffResponse:
        parsed_request = parse_diff_request(self, request, context)
        _parsed, repo, ctx_from, ctx_to = parsed_request
        if not _parsed:
            context.abort(StatusCode.INTERNAL, "exit status 128")
        opts = {b'git': True}
        overrides = {
            (b'experimental', b'extendedheader.similarity'): True,
        }
        with repo.ui.configoverride(overrides):
            diffopts = diffutil.diffallopts(repo.ui, opts)
            diffchunks = ctx_to.diff(ctx_from, opts=diffopts)

        # generator func to yield hunks
        def in_chunks():
            for hg_chunk in patchmod.parsepatch(diffchunks):
                chunk = Chunk(hg_chunk, ctx_from, ctx_to)
                header, _bin_placeholder = chunk.header_with_index_line()
                yield header

                for hunk in hg_chunk.hunks:
                    with BytesIO() as extracted:
                        hunk.write(extracted)
                        yield extracted.getvalue()
        for data in concat_resplit(in_chunks(), WRITE_BUFFER_SIZE):
            yield RawDiffResponse(data=data)

    def RawPatch(self, request: RawPatchRequest,
                 context) -> RawPatchResponse:
        """Yields raw patches between two csets.

        Note: Here, patches are in `hg` format instead of `git`. We decided
        this on the basis of "response is not being parsed at Rails side, and
        directly sent to UI" so that users can import the hg patches.
        """
        parsed_request = parse_diff_request(self, request, context)
        _parsed, repo, ctx_from, ctx_to = parsed_request
        if not _parsed:
            context.abort(StatusCode.INTERNAL, "exit status 128")
        opts = {b'git': True}
        diffopts = diffutil.diffallopts(repo.ui, opts)
        ui = repo.ui
        fm = ui.formatter(b'RawPatch', opts={})
        revs = repo.revs(b'only(%s, %s)', ctx_to, ctx_from)

        def in_chunks():
            for seqno, rev in enumerate(revs):
                ctx = repo[rev]
                itr_data = _exportsingle(repo, ctx, fm, seqno, diffopts)
                for data in itr_data:
                    yield data
        for data in concat_resplit(in_chunks(), WRITE_BUFFER_SIZE):
            yield RawPatchResponse(data=data)

    def DiffStats(self, request: DiffStatsRequest,
                  context) -> DiffStatsResponse:
        parsed_request = parse_diff_request(self, request, context)
        _parsed, repo, ctx_from, ctx_to = parsed_request
        if not _parsed:
            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")

        diffchunks = ctx_to.diff(ctx_from, opts=diff_opts(repo))

        for stats in aggregate_flush_batch(
                chunk_stats(diffchunks, ctx_from, ctx_to),
                lambda x: 1,
                MAX_NUM_STAT_BATCH_SIZE):
            yield DiffStatsResponse(stats=stats)

    def FindChangedPaths(self, request: FindChangedPathsRequest,
                         context) -> FindChangedPathsResponse:
        repo = self.load_repo(request.repository, context)
        extracted = []  # triplets (changeset1, changeset2, path)
        SubReq = FindChangedPathsRequest.Request
        CommitRequest = SubReq.CommitRequest

        sub_reqs = request.requests
        if request.commits:
            # compat wrapper for deprecated call style. Quoting diff.proto:
            #   This field is deprecated. To adapt to the new calling
            #   convention you can create one `CommitRequest` per commit,
            #   where each `CommitRequest` has only the `commit_revision`
            sub_reqs.extend(
                SubReq(commit_request=CommitRequest(commit_revision=rev))
                for rev in request.commits
            )

        # Gitaly (as of gitaly@2b069d853) simply issues a `git diff-tree`
        # per sub-request and yields the responses of each.
        # In other words, there is no merging of results. For instance,
        # with two sub-requests, a file appearing modified for both of them
        # will give rise to two identical ChangedPaths messages.
        for sub_req in sub_reqs:
            if sub_req.HasField('commit_request'):
                right_ctx = fcp_resolve_commit(
                    context, repo,
                    sub_req.commit_request.commit_revision)

                parent_revs = sub_req.commit_request.parent_commit_revisions
                if parent_revs:
                    parents = [fcp_resolve_commit(context, repo, rev)
                               for rev in parent_revs]
                else:
                    parents = right_ctx.parents()

                extracted.extend((left_ctx, right_ctx, None)
                                 for left_ctx in parents)
            else:
                # TODO trees can probably be specified in some arcane Git
                # traversal rather than just by oid.
                oids = [sub_req.tree_request.left_tree_revision,
                        sub_req.tree_request.right_tree_revision]
                cids_paths = [split_chgsid_path(oid) for oid in oids]
                path = cids_paths[0][1]
                if cids_paths[1][1] != path:
                    context.abort(StatusCode.UNIMPLEMENTED,
                                  "Cannot compare yet with different paths. "
                                  "Got TreeRequest with left and right "
                                  "resolving as %r" % cids_paths)
                unfi = repo.unfiltered()
                left_ctx, right_ctx = [unfi[cp[0].encode('ascii')]
                                       for cp in cids_paths]
                extracted.append((left_ctx, right_ctx, path))

        for paths in chunked(path for extr in extracted
                             for path in changed_paths(repo, *extr)):
            yield FindChangedPathsResponse(paths=paths)

    def GetPatchID(self, request: GetPatchIDRequest,
                   context) -> GetPatchIDResponse:
        repo = self.load_repo(request.repository, context)
        old_changeset = gitlab_revision_changeset(repo, request.old_revision)
        if old_changeset is None:
            context.abort(StatusCode.INTERNAL,
                          "revision %r not found" % request.old_revision)
        new_changeset = gitlab_revision_changeset(repo, request.new_revision)
        if new_changeset is None:
            context.abort(StatusCode.INTERNAL,
                          "revision %r not found" % request.new_revision)

        git_path = repo.ui.config(b'hgitaly', b'git-executable')

        try:
            git_out = git_patch_id(git_path, old_changeset, new_changeset)
        except FileNotFoundError:
            context.abort(StatusCode.INTERNAL,
                          "Expected Git executable not found at %r" % git_path)
        except PermissionError:
            context.abort(StatusCode.INTERNAL,
                          "Expected Git executable found at %r, but it is "
                          "not executable" % git_path)

        # like Gitaly, ignoring the second hash, which is useful only
        # in Git tree comparisons (to recall a Git commit id)
        return GetPatchIDResponse(patch_id=git_out.split()[0])


def fcp_resolve_commit(context, repo, revision):
    ctx = gitlab_revision_changeset(
        repo, parse_diff_request_cid(revision))
    if ctx is None:
        context.abort(
            StatusCode.NOT_FOUND,
            'resolving commit: revision can not be found: '
            '"%s"' % pycompat.sysstr(revision)
        )
    return ctx


def parse_diff_request_cid(cid):
    """Perform the conversions from a request commit_id to a usable revision.
    """
    if cid == EMPTY_TREE_OID:
        return nullhex
    return pycompat.sysbytes(cid)


def parse_diff_request(servicer, request, context):
    logger = LoggerAdapter(base_logger, context)
    repo = servicer.load_repo(request.repository, context)

    left_cid = parse_diff_request_cid(request.left_commit_id)
    right_cid = parse_diff_request_cid(request.right_commit_id)

    ctx_from = gitlab_revision_changeset(repo, left_cid)
    ctx_to = gitlab_revision_changeset(repo, right_cid)
    if ctx_from is None:
        logger.warning(
            "%s: left_commit_id %r "
            "could not be found", request.__class__.__name__, left_cid)
        return (False, repo, ctx_from, ctx_to)
    if ctx_to is None:
        logger.warning(
            "%s: right_commit_id %r "
            "could not be found", request.__class__.__name__, right_cid)
        return (False, repo, ctx_from, ctx_to)
    return (True, repo, ctx_from, ctx_to)


def _exportsingle(repo, ctx, fm, seqno, diffopts):
    """Generator method which yields a bytes stream of exporting `ctx` data.

    This method overwrite upstream mercurial's cmdutil._exportsingle(), as
    the upstream version directly writes the data to stdout and concatenates
    the diff chunks instead of yielding them.
    """
    node = ctx.node()
    parents = [p.node() for p in ctx.parents() if p]
    branch = ctx.branch()

    if parents:
        p1 = parents[0]
    else:
        p1 = nullid

    textlines = []
    textlines.append(b'# HG changeset patch\n')
    textlines.append(b'# User %s\n' % ctx.user())
    textlines.append(b'# Date %d %d\n' % ctx.date())
    textlines.append(b'#      %s\n' % fm.formatdate(ctx.date()))
    if branch and branch != b'default':
        textlines.append(b'# Branch %s\n' % branch)
    textlines.append(b'# Node ID %s\n' % hex(node))
    textlines.append(b'# Parent  %s\n' % hex(p1))
    if len(parents) > 1:
        textlines.append(b'# Parent  %s\n' % hex(parents[1]))

    for headerid in cmdutil.extraexport:
        header = cmdutil.extraexportmap[headerid](seqno, ctx)
        if header is not None:
            textlines.append(b'# %s\n' % header)

    textlines.append(b'%s\n' % ctx.description().rstrip())
    textlines.append(b'\n')
    yield b''.join(textlines)

    chunkiter = patchmod.diff(repo, p1, node, opts=diffopts)
    for chunk in chunkiter:
        yield chunk


def commit_delta_size(delta):
    size = (
        len(delta.from_id) + len(delta.to_id)
        + 4 + 4  # old_mode and new_mode are int, each 4 bytes
        + len(delta.from_path) + len(delta.to_path)
    )
    return size



from pathlib import Path
from setuptools import setup, find_packages

with open('install-requirements.txt', 'r') as install_reqf:
    install_req = [req.strip() for req in install_reqf]

VERSION_FILE = 'VERSION'
MAIN_PACKAGE = 'hgitaly'

scripts = {  # name to (module in main package, function)
    'hgitaly-stats': ('scripts', 'stats'),
}

setup(
    name='hgitaly',
    version=Path('hgitaly', VERSION_FILE).read_text().strip(),
    author='Georges Racinet',
    author_email='georges.racinet@octobus.net',
    url='https://foss.heptapod.net/heptapod/hgitaly',
    description="Server-side implementation of Gitaly protocol for Mercurial",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='hg mercurial heptapod gitlab',
    license='GPLv2+',
    packages=find_packages(),
    package_data={'hgitaly': [VERSION_FILE],
                  'hgitaly.linguist': ['languages.json'],
                  'hgitaly.license_detector': ['spdx-licenses.json'],
                  },
    entry_points=dict(
        console_scripts=[
            '{name}={pkg}.{mod}:{fun}'.format(
                pkg=MAIN_PACKAGE, name=name, mod=mod, fun=fun)
            for name, (mod, fun) in scripts.items()],
    ),
    install_requires=install_req,
    python_requires='>=3.8',
)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
from importlib import reload

from mercurial import sshpeer

from mercurial_testhelpers.repo_wrapper import RepoWrapper
from ..sshd import hg_ssh_setup
from hgitaly.ssh import client_command_str
from hgitaly.testing.ssh import hg_exe_path


@pytest.fixture
def ssh_setup(tmpdir):
    working_dir = tmpdir / 'working'
    working_dir.mkdir()
    yield from hg_ssh_setup(working_dir)


def test_ssh_server(ssh_setup, tmpdir):
    server, client_key, known_hosts = ssh_setup
    src = RepoWrapper.init(tmpdir / 'src')
    dest = RepoWrapper.init(tmpdir / 'dst')
    ctx = src.commit_file('foo')
    reload(sshpeer)

    with client_command_str(client_key, known_hosts) as ssh_cmd:
        src.command('push',
                    f'ssh://{server.host}:{server.port}/{dest.path}'.encode(),
                    ssh=ssh_cmd.encode(),
                    remotecmd=hg_exe_path().encode(),
                    )
    assert ctx in dest.repo



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import grpc
import pytest
from mercurial import pycompat

from hgitaly.tests.common import make_empty_repo

from hgitaly.stub.ref_pb2 import (
    FindDefaultBranchNameRequest,
    FindDefaultBranchNameResponse
)
from hgitaly.stub.ref_pb2_grpc import RefServiceStub
from hgitaly.stub.repository_pb2 import WriteRefRequest
from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub
from ..ref import DEFAULT_BRANCH_FILE_NAME

DEFAULT_BRANCH_FILE_NAME = pycompat.sysstr(DEFAULT_BRANCH_FILE_NAME)


def test_default_branch(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    repo_stub = RepositoryServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    request = FindDefaultBranchNameRequest(repository=grpc_repo)
    response = ref_stub.FindDefaultBranchName(request)
    assert response == FindDefaultBranchNameResponse()

    wreq = WriteRefRequest(repository=grpc_repo,
                           ref=b'HEAD',
                           revision=b'refs/heads/branch/other')
    repo_stub.WriteRef(wreq)

    request = FindDefaultBranchNameRequest(repository=grpc_repo)
    response = ref_stub.FindDefaultBranchName(request)
    # branch does not exist yet, hence gets filtered out
    assert not response.name

    wrapper.commit_file('foo', branch='other')

    request = FindDefaultBranchNameRequest(repository=grpc_repo)
    response = ref_stub.FindDefaultBranchName(request)
    assert response.name == b'refs/heads/branch/other'

    grpc_repo.storage_name = 'unknown'
    request = FindDefaultBranchNameRequest(repository=grpc_repo)
    with pytest.raises(grpc.RpcError) as exc_info:
        ref_stub.FindDefaultBranchName(request)
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: commit.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import errors_pb2 as errors__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import lint_pb2 as lint__pb2
from . import shared_pb2 as shared__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63ommit.proto\x12\x06gitaly\x1a\x0c\x65rrors.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\nlint.proto\x1a\x0cshared.proto\"\xd4\x03\n\x12ListCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x11\n\trevisions\x18\x02 \x03(\t\x12\x36\n\x11pagination_params\x18\x03 \x01(\x0b\x32\x1b.gitaly.PaginationParameter\x12/\n\x05order\x18\x04 \x01(\x0e\x32 .gitaly.ListCommitsRequest.Order\x12\x0f\n\x07reverse\x18\x0b \x01(\x08\x12\x13\n\x0bmax_parents\x18\x05 \x01(\r\x12\x14\n\x0c\x64isable_walk\x18\x06 \x01(\x08\x12\x14\n\x0c\x66irst_parent\x18\x07 \x01(\x08\x12)\n\x05\x61\x66ter\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06\x62\x65\x66ore\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06\x61uthor\x18\n \x01(\x0c\x12\x13\n\x0bignore_case\x18\x0c \x01(\x08\x12\x1f\n\x17\x63ommit_message_patterns\x18\r \x03(\x0c\"%\n\x05Order\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04TOPO\x10\x01\x12\x08\n\x04\x44\x41TE\x10\x02\"9\n\x13ListCommitsResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"}\n\x15ListAllCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x36\n\x11pagination_params\x18\x02 \x01(\x0b\x32\x1b.gitaly.PaginationParameter\"<\n\x16ListAllCommitsResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"T\n\x12\x43ommitStatsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\"H\n\x13\x43ommitStatsResponse\x12\x0b\n\x03oid\x18\x01 \x01(\t\x12\x11\n\tadditions\x18\x02 \x01(\x05\x12\x11\n\tdeletions\x18\x03 \x01(\x05\"n\n\x17\x43ommitIsAncestorRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x13\n\x0b\x61ncestor_id\x18\x02 \x01(\t\x12\x10\n\x08\x63hild_id\x18\x03 \x01(\t\")\n\x18\x43ommitIsAncestorResponse\x12\r\n\x05value\x18\x01 \x01(\x08\"\x81\x01\n\x10TreeEntryRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12\r\n\x05limit\x18\x04 \x01(\x03\x12\x10\n\x08max_size\x18\x05 \x01(\x03\"\xb5\x01\n\x11TreeEntryResponse\x12\x32\n\x04type\x18\x01 \x01(\x0e\x32$.gitaly.TreeEntryResponse.ObjectType\x12\x0b\n\x03oid\x18\x02 \x01(\t\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x0c\n\x04mode\x18\x04 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\"5\n\nObjectType\x12\n\n\x06\x43OMMIT\x10\x00\x12\x08\n\x04\x42LOB\x10\x01\x12\x08\n\x04TREE\x10\x02\x12\x07\n\x03TAG\x10\x03\"\x9f\x02\n\x13\x43ountCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12)\n\x05\x61\x66ter\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06\x62\x65\x66ore\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04path\x18\x05 \x01(\x0c\x12\x11\n\tmax_count\x18\x06 \x01(\x05\x12\x0b\n\x03\x61ll\x18\x07 \x01(\x08\x12\x14\n\x0c\x66irst_parent\x18\x08 \x01(\x08\x12-\n\x0eglobal_options\x18\t \x01(\x0b\x32\x15.gitaly.GlobalOptions\"%\n\x14\x43ountCommitsResponse\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\"\x8b\x01\n\x1c\x43ountDivergingCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x0c\n\x04\x66rom\x18\x02 \x01(\x0c\x12\n\n\x02to\x18\x03 \x01(\x0c\x12\x11\n\tmax_count\x18\x07 \x01(\x05J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07\"H\n\x1d\x43ountDivergingCommitsResponse\x12\x12\n\nleft_count\x18\x01 \x01(\x05\x12\x13\n\x0bright_count\x18\x02 \x01(\x05\"\xc3\x01\n\tTreeEntry\x12\x0b\n\x03oid\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12)\n\x04type\x18\x04 \x01(\x0e\x32\x1b.gitaly.TreeEntry.EntryType\x12\x0c\n\x04mode\x18\x05 \x01(\x05\x12\x12\n\ncommit_oid\x18\x06 \x01(\t\x12\x11\n\tflat_path\x18\x07 \x01(\x0c\"+\n\tEntryType\x12\x08\n\x04\x42LOB\x10\x00\x12\x08\n\x04TREE\x10\x01\x12\n\n\x06\x43OMMIT\x10\x03J\x04\x08\x02\x10\x03R\x08root_oid\"\xa5\x02\n\x15GetTreeEntriesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12\x11\n\trecursive\x18\x04 \x01(\x08\x12\x32\n\x04sort\x18\x05 \x01(\x0e\x32$.gitaly.GetTreeEntriesRequest.SortBy\x12\x36\n\x11pagination_params\x18\x06 \x01(\x0b\x32\x1b.gitaly.PaginationParameter\x12\x17\n\x0fskip_flat_paths\x18\x07 \x01(\x08\"&\n\x06SortBy\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0f\n\x0bTREES_FIRST\x10\x01\"q\n\x16GetTreeEntriesResponse\x12\"\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x11.gitaly.TreeEntry\x12\x33\n\x11pagination_cursor\x18\x02 \x01(\x0b\x32\x18.gitaly.PaginationCursor\"w\n\x13GetTreeEntriesError\x12\x34\n\x0cresolve_tree\x18\x01 \x01(\x0b\x32\x1c.gitaly.ResolveRevisionErrorH\x00\x12!\n\x04path\x18\x02 \x01(\x0b\x32\x11.gitaly.PathErrorH\x00\x42\x07\n\x05\x65rror\"R\n\x10ListFilesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\"\"\n\x11ListFilesResponse\x12\r\n\x05paths\x18\x01 \x03(\x0c\"e\n\x11\x46indCommitRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x10\n\x08trailers\x18\x03 \x01(\x08\"7\n\x12\x46indCommitResponse\x12!\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x11.gitaly.GitCommit\"T\n\x17ListCommitsByOidRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x0b\n\x03oid\x18\x02 \x03(\t\">\n\x18ListCommitsByOidResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"^\n\x1bListCommitsByRefNameRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x11\n\tref_names\x18\x02 \x03(\x0c\"\xb1\x01\n\x1cListCommitsByRefNameResponse\x12\x46\n\x0b\x63ommit_refs\x18\x02 \x03(\x0b\x32\x31.gitaly.ListCommitsByRefNameResponse.CommitForRef\x1a\x43\n\x0c\x43ommitForRef\x12!\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x11.gitaly.GitCommit\x12\x10\n\x08ref_name\x18\x02 \x01(\x0cJ\x04\x08\x01\x10\x02\"\xd3\x01\n\x15\x46indAllCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x11\n\tmax_count\x18\x03 \x01(\x05\x12\x0c\n\x04skip\x18\x04 \x01(\x05\x12\x32\n\x05order\x18\x05 \x01(\x0e\x32#.gitaly.FindAllCommitsRequest.Order\"%\n\x05Order\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04TOPO\x10\x01\x12\x08\n\x04\x44\x41TE\x10\x02\"<\n\x16\x46indAllCommitsResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"\x94\x04\n\x12\x46indCommitsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x0e\n\x06offset\x18\x04 \x01(\x05\x12\r\n\x05paths\x18\x05 \x03(\x0c\x12\x0e\n\x06\x66ollow\x18\x06 \x01(\x08\x12\x13\n\x0bskip_merges\x18\x07 \x01(\x08\x12\x18\n\x0c\x64isable_walk\x18\x08 \x01(\x08\x42\x02\x18\x01\x12)\n\x05\x61\x66ter\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06\x62\x65\x66ore\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0b\n\x03\x61ll\x18\x0b \x01(\x08\x12\x14\n\x0c\x66irst_parent\x18\x0c \x01(\x08\x12\x0e\n\x06\x61uthor\x18\r \x01(\x0c\x12/\n\x05order\x18\x0e \x01(\x0e\x32 .gitaly.FindCommitsRequest.Order\x12-\n\x0eglobal_options\x18\x0f \x01(\x0b\x32\x15.gitaly.GlobalOptions\x12\x10\n\x08trailers\x18\x10 \x01(\x08\x12\x19\n\x11include_shortstat\x18\x11 \x01(\x08\x12\x1d\n\x15include_referenced_by\x18\x12 \x03(\x0c\"\x1b\n\x05Order\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04TOPO\x10\x01\"9\n\x13\x46indCommitsResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"\xba\x01\n\x10\x46indCommitsError\x12,\n\nbad_object\x18\x01 \x01(\x0b\x32\x16.gitaly.BadObjectErrorH\x00\x12\x38\n\rambiguous_ref\x18\x02 \x01(\x0b\x32\x1f.gitaly.AmbiguousReferenceErrorH\x00\x12\x35\n\rinvalid_range\x18\x03 \x01(\x0b\x32\x1c.gitaly.InvalidRevisionRangeH\x00\x42\x07\n\x05\x65rror\"X\n\x16\x43ommitLanguagesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\"\xaf\x01\n\x17\x43ommitLanguagesResponse\x12;\n\tlanguages\x18\x01 \x03(\x0b\x32(.gitaly.CommitLanguagesResponse.Language\x1aW\n\x08Language\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05share\x18\x02 \x01(\x02\x12\r\n\x05\x63olor\x18\x03 \x01(\t\x12\r\n\x05\x62ytes\x18\x05 \x01(\x04J\x04\x08\x04\x10\x05R\nfile_count\"n\n\x0fRawBlameRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12\r\n\x05range\x18\x04 \x01(\x0c\" \n\x10RawBlameResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\xb5\x01\n\rRawBlameError\x12\x33\n\x0epath_not_found\x18\x01 \x01(\x0b\x32\x19.gitaly.PathNotFoundErrorH\x00\x12=\n\x0cout_of_range\x18\x02 \x01(\x0b\x32%.gitaly.RawBlameError.OutOfRangeErrorH\x00\x1a\'\n\x0fOutOfRangeError\x12\x14\n\x0c\x61\x63tual_lines\x18\x01 \x01(\x04\x42\x07\n\x05\x65rror\"\xb1\x01\n\x18LastCommitForPathRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12\x18\n\x10literal_pathspec\x18\x04 \x01(\x08\x12-\n\x0eglobal_options\x18\x05 \x01(\x0b\x32\x15.gitaly.GlobalOptions\">\n\x19LastCommitForPathResponse\x12!\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x11.gitaly.GitCommit\"\xd9\x01\n\x1dListLastCommitsForTreeRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\x0c\x12\r\n\x05limit\x18\x04 \x01(\x05\x12\x0e\n\x06offset\x18\x05 \x01(\x05\x12\x1c\n\x10literal_pathspec\x18\x06 \x01(\x08\x42\x02\x18\x01\x12-\n\x0eglobal_options\x18\x07 \x01(\x0b\x32\x15.gitaly.GlobalOptions\"\xbb\x01\n\x1eListLastCommitsForTreeResponse\x12\x45\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x34.gitaly.ListLastCommitsForTreeResponse.CommitForTree\x1aR\n\rCommitForTree\x12!\n\x06\x63ommit\x18\x02 \x01(\x0b\x32\x11.gitaly.GitCommit\x12\x12\n\npath_bytes\x18\x04 \x01(\x0cJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04\"\xc4\x01\n\x17\x43ommitsByMessageRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08revision\x18\x02 \x01(\x0c\x12\x0e\n\x06offset\x18\x03 \x01(\x05\x12\r\n\x05limit\x18\x04 \x01(\x05\x12\x0c\n\x04path\x18\x05 \x01(\x0c\x12\r\n\x05query\x18\x06 \x01(\t\x12-\n\x0eglobal_options\x18\x07 \x01(\x0b\x32\x15.gitaly.GlobalOptions\">\n\x18\x43ommitsByMessageResponse\x12\"\n\x07\x63ommits\x18\x01 \x03(\x0b\x32\x11.gitaly.GitCommit\"]\n\x1f\x46ilterShasWithSignaturesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x0c\n\x04shas\x18\x02 \x03(\x0c\"0\n FilterShasWithSignaturesResponse\x12\x0c\n\x04shas\x18\x01 \x03(\x0c\"^\n\x1aGetCommitSignaturesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x12\n\ncommit_ids\x18\x02 \x03(\t\"\xda\x01\n\x1bGetCommitSignaturesResponse\x12\x11\n\tcommit_id\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\x13\n\x0bsigned_text\x18\x03 \x01(\x0c\x12:\n\x06signer\x18\x04 \x01(\x0e\x32*.gitaly.GetCommitSignaturesResponse.Signer\"D\n\x06Signer\x12\x16\n\x12SIGNER_UNSPECIFIED\x10\x00\x12\x0f\n\x0bSIGNER_USER\x10\x01\x12\x11\n\rSIGNER_SYSTEM\x10\x02\"\\\n\x18GetCommitMessagesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x12\n\ncommit_ids\x18\x02 \x03(\t\"?\n\x19GetCommitMessagesResponse\x12\x11\n\tcommit_id\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\x0c\"[\n\x18\x43heckObjectsExistRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x11\n\trevisions\x18\x02 \x03(\x0c\"\x96\x01\n\x19\x43heckObjectsExistResponse\x12\x46\n\trevisions\x18\x01 \x03(\x0b\x32\x33.gitaly.CheckObjectsExistResponse.RevisionExistence\x1a\x31\n\x11RevisionExistence\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\x0e\n\x06\x65xists\x18\x02 \x01(\x08\x32\xf1\x10\n\rCommitService\x12P\n\x0bListCommits\x12\x1a.gitaly.ListCommitsRequest\x1a\x1b.gitaly.ListCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12Y\n\x0eListAllCommits\x12\x1d.gitaly.ListAllCommitsRequest\x1a\x1e.gitaly.ListAllCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12]\n\x10\x43ommitIsAncestor\x12\x1f.gitaly.CommitIsAncestorRequest\x1a .gitaly.CommitIsAncestorResponse\"\x06\xfa\x97(\x02\x08\x02\x12J\n\tTreeEntry\x12\x18.gitaly.TreeEntryRequest\x1a\x19.gitaly.TreeEntryResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12Q\n\x0c\x43ountCommits\x12\x1b.gitaly.CountCommitsRequest\x1a\x1c.gitaly.CountCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x12l\n\x15\x43ountDivergingCommits\x12$.gitaly.CountDivergingCommitsRequest\x1a%.gitaly.CountDivergingCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x12Y\n\x0eGetTreeEntries\x12\x1d.gitaly.GetTreeEntriesRequest\x1a\x1e.gitaly.GetTreeEntriesResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12J\n\tListFiles\x12\x18.gitaly.ListFilesRequest\x1a\x19.gitaly.ListFilesResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12K\n\nFindCommit\x12\x19.gitaly.FindCommitRequest\x1a\x1a.gitaly.FindCommitResponse\"\x06\xfa\x97(\x02\x08\x02\x12N\n\x0b\x43ommitStats\x12\x1a.gitaly.CommitStatsRequest\x1a\x1b.gitaly.CommitStatsResponse\"\x06\xfa\x97(\x02\x08\x02\x12Y\n\x0e\x46indAllCommits\x12\x1d.gitaly.FindAllCommitsRequest\x1a\x1e.gitaly.FindAllCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12P\n\x0b\x46indCommits\x12\x1a.gitaly.FindCommitsRequest\x1a\x1b.gitaly.FindCommitsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12Z\n\x0f\x43ommitLanguages\x12\x1e.gitaly.CommitLanguagesRequest\x1a\x1f.gitaly.CommitLanguagesResponse\"\x06\xfa\x97(\x02\x08\x02\x12G\n\x08RawBlame\x12\x17.gitaly.RawBlameRequest\x1a\x18.gitaly.RawBlameResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12`\n\x11LastCommitForPath\x12 .gitaly.LastCommitForPathRequest\x1a!.gitaly.LastCommitForPathResponse\"\x06\xfa\x97(\x02\x08\x02\x12q\n\x16ListLastCommitsForTree\x12%.gitaly.ListLastCommitsForTreeRequest\x1a&.gitaly.ListLastCommitsForTreeResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12_\n\x10\x43ommitsByMessage\x12\x1f.gitaly.CommitsByMessageRequest\x1a .gitaly.CommitsByMessageResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12_\n\x10ListCommitsByOid\x12\x1f.gitaly.ListCommitsByOidRequest\x1a .gitaly.ListCommitsByOidResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12k\n\x14ListCommitsByRefName\x12#.gitaly.ListCommitsByRefNameRequest\x1a$.gitaly.ListCommitsByRefNameResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12y\n\x18\x46ilterShasWithSignatures\x12\'.gitaly.FilterShasWithSignaturesRequest\x1a(.gitaly.FilterShasWithSignaturesResponse\"\x06\xfa\x97(\x02\x08\x02(\x01\x30\x01\x12h\n\x13GetCommitSignatures\x12\".gitaly.GetCommitSignaturesRequest\x1a#.gitaly.GetCommitSignaturesResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12\x62\n\x11GetCommitMessages\x12 .gitaly.GetCommitMessagesRequest\x1a!.gitaly.GetCommitMessagesResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12\x64\n\x11\x43heckObjectsExist\x12 .gitaly.CheckObjectsExistRequest\x1a!.gitaly.CheckObjectsExistResponse\"\x06\xfa\x97(\x02\x08\x02(\x01\x30\x01\x42\x34Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'commit_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb'
  _LISTCOMMITSREQUEST.fields_by_name['repository']._options = None
  _LISTCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTALLCOMMITSREQUEST.fields_by_name['repository']._options = None
  _LISTALLCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _COMMITSTATSREQUEST.fields_by_name['repository']._options = None
  _COMMITSTATSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _COMMITISANCESTORREQUEST.fields_by_name['repository']._options = None
  _COMMITISANCESTORREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _TREEENTRYREQUEST.fields_by_name['repository']._options = None
  _TREEENTRYREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _COUNTCOMMITSREQUEST.fields_by_name['repository']._options = None
  _COUNTCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _COUNTDIVERGINGCOMMITSREQUEST.fields_by_name['repository']._options = None
  _COUNTDIVERGINGCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETTREEENTRIESREQUEST.fields_by_name['repository']._options = None
  _GETTREEENTRIESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTFILESREQUEST.fields_by_name['repository']._options = None
  _LISTFILESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _FINDCOMMITREQUEST.fields_by_name['repository']._options = None
  _FINDCOMMITREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTCOMMITSBYOIDREQUEST.fields_by_name['repository']._options = None
  _LISTCOMMITSBYOIDREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTCOMMITSBYREFNAMEREQUEST.fields_by_name['repository']._options = None
  _LISTCOMMITSBYREFNAMEREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _FINDALLCOMMITSREQUEST.fields_by_name['repository']._options = None
  _FINDALLCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _FINDCOMMITSREQUEST.fields_by_name['repository']._options = None
  _FINDCOMMITSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _FINDCOMMITSREQUEST.fields_by_name['disable_walk']._options = None
  _FINDCOMMITSREQUEST.fields_by_name['disable_walk']._serialized_options = b'\030\001'
  _COMMITLANGUAGESREQUEST.fields_by_name['repository']._options = None
  _COMMITLANGUAGESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _RAWBLAMEREQUEST.fields_by_name['repository']._options = None
  _RAWBLAMEREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LASTCOMMITFORPATHREQUEST.fields_by_name['repository']._options = None
  _LASTCOMMITFORPATHREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTLASTCOMMITSFORTREEREQUEST.fields_by_name['repository']._options = None
  _LISTLASTCOMMITSFORTREEREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTLASTCOMMITSFORTREEREQUEST.fields_by_name['literal_pathspec']._options = None
  _LISTLASTCOMMITSFORTREEREQUEST.fields_by_name['literal_pathspec']._serialized_options = b'\030\001'
  _COMMITSBYMESSAGEREQUEST.fields_by_name['repository']._options = None
  _COMMITSBYMESSAGEREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _FILTERSHASWITHSIGNATURESREQUEST.fields_by_name['repository']._options = None
  _FILTERSHASWITHSIGNATURESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETCOMMITSIGNATURESREQUEST.fields_by_name['repository']._options = None
  _GETCOMMITSIGNATURESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETCOMMITMESSAGESREQUEST.fields_by_name['repository']._options = None
  _GETCOMMITMESSAGESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _CHECKOBJECTSEXISTREQUEST.fields_by_name['repository']._options = None
  _CHECKOBJECTSEXISTREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _COMMITSERVICE.methods_by_name['ListCommits']._options = None
  _COMMITSERVICE.methods_by_name['ListCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['ListAllCommits']._options = None
  _COMMITSERVICE.methods_by_name['ListAllCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CommitIsAncestor']._options = None
  _COMMITSERVICE.methods_by_name['CommitIsAncestor']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['TreeEntry']._options = None
  _COMMITSERVICE.methods_by_name['TreeEntry']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CountCommits']._options = None
  _COMMITSERVICE.methods_by_name['CountCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CountDivergingCommits']._options = None
  _COMMITSERVICE.methods_by_name['CountDivergingCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['GetTreeEntries']._options = None
  _COMMITSERVICE.methods_by_name['GetTreeEntries']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['ListFiles']._options = None
  _COMMITSERVICE.methods_by_name['ListFiles']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['FindCommit']._options = None
  _COMMITSERVICE.methods_by_name['FindCommit']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CommitStats']._options = None
  _COMMITSERVICE.methods_by_name['CommitStats']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['FindAllCommits']._options = None
  _COMMITSERVICE.methods_by_name['FindAllCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['FindCommits']._options = None
  _COMMITSERVICE.methods_by_name['FindCommits']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CommitLanguages']._options = None
  _COMMITSERVICE.methods_by_name['CommitLanguages']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['RawBlame']._options = None
  _COMMITSERVICE.methods_by_name['RawBlame']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['LastCommitForPath']._options = None
  _COMMITSERVICE.methods_by_name['LastCommitForPath']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['ListLastCommitsForTree']._options = None
  _COMMITSERVICE.methods_by_name['ListLastCommitsForTree']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CommitsByMessage']._options = None
  _COMMITSERVICE.methods_by_name['CommitsByMessage']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['ListCommitsByOid']._options = None
  _COMMITSERVICE.methods_by_name['ListCommitsByOid']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['ListCommitsByRefName']._options = None
  _COMMITSERVICE.methods_by_name['ListCommitsByRefName']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['FilterShasWithSignatures']._options = None
  _COMMITSERVICE.methods_by_name['FilterShasWithSignatures']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['GetCommitSignatures']._options = None
  _COMMITSERVICE.methods_by_name['GetCommitSignatures']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['GetCommitMessages']._options = None
  _COMMITSERVICE.methods_by_name['GetCommitMessages']._serialized_options = b'\372\227(\002\010\002'
  _COMMITSERVICE.methods_by_name['CheckObjectsExist']._options = None
  _COMMITSERVICE.methods_by_name['CheckObjectsExist']._serialized_options = b'\372\227(\002\010\002'
  _globals['_LISTCOMMITSREQUEST']._serialized_start=98
  _globals['_LISTCOMMITSREQUEST']._serialized_end=566
  _globals['_LISTCOMMITSREQUEST_ORDER']._serialized_start=529
  _globals['_LISTCOMMITSREQUEST_ORDER']._serialized_end=566
  _globals['_LISTCOMMITSRESPONSE']._serialized_start=568
  _globals['_LISTCOMMITSRESPONSE']._serialized_end=625
  _globals['_LISTALLCOMMITSREQUEST']._serialized_start=627
  _globals['_LISTALLCOMMITSREQUEST']._serialized_end=752
  _globals['_LISTALLCOMMITSRESPONSE']._serialized_start=754
  _globals['_LISTALLCOMMITSRESPONSE']._serialized_end=814
  _globals['_COMMITSTATSREQUEST']._serialized_start=816
  _globals['_COMMITSTATSREQUEST']._serialized_end=900
  _globals['_COMMITSTATSRESPONSE']._serialized_start=902
  _globals['_COMMITSTATSRESPONSE']._serialized_end=974
  _globals['_COMMITISANCESTORREQUEST']._serialized_start=976
  _globals['_COMMITISANCESTORREQUEST']._serialized_end=1086
  _globals['_COMMITISANCESTORRESPONSE']._serialized_start=1088
  _globals['_COMMITISANCESTORRESPONSE']._serialized_end=1129
  _globals['_TREEENTRYREQUEST']._serialized_start=1132
  _globals['_TREEENTRYREQUEST']._serialized_end=1261
  _globals['_TREEENTRYRESPONSE']._serialized_start=1264
  _globals['_TREEENTRYRESPONSE']._serialized_end=1445
  _globals['_TREEENTRYRESPONSE_OBJECTTYPE']._serialized_start=1392
  _globals['_TREEENTRYRESPONSE_OBJECTTYPE']._serialized_end=1445
  _globals['_COUNTCOMMITSREQUEST']._serialized_start=1448
  _globals['_COUNTCOMMITSREQUEST']._serialized_end=1735
  _globals['_COUNTCOMMITSRESPONSE']._serialized_start=1737
  _globals['_COUNTCOMMITSRESPONSE']._serialized_end=1774
  _globals['_COUNTDIVERGINGCOMMITSREQUEST']._serialized_start=1777
  _globals['_COUNTDIVERGINGCOMMITSREQUEST']._serialized_end=1916
  _globals['_COUNTDIVERGINGCOMMITSRESPONSE']._serialized_start=1918
  _globals['_COUNTDIVERGINGCOMMITSRESPONSE']._serialized_end=1990
  _globals['_TREEENTRY']._serialized_start=1993
  _globals['_TREEENTRY']._serialized_end=2188
  _globals['_TREEENTRY_ENTRYTYPE']._serialized_start=2129
  _globals['_TREEENTRY_ENTRYTYPE']._serialized_end=2172
  _globals['_GETTREEENTRIESREQUEST']._serialized_start=2191
  _globals['_GETTREEENTRIESREQUEST']._serialized_end=2484
  _globals['_GETTREEENTRIESREQUEST_SORTBY']._serialized_start=2446
  _globals['_GETTREEENTRIESREQUEST_SORTBY']._serialized_end=2484
  _globals['_GETTREEENTRIESRESPONSE']._serialized_start=2486
  _globals['_GETTREEENTRIESRESPONSE']._serialized_end=2599
  _globals['_GETTREEENTRIESERROR']._serialized_start=2601
  _globals['_GETTREEENTRIESERROR']._serialized_end=2720
  _globals['_LISTFILESREQUEST']._serialized_start=2722
  _globals['_LISTFILESREQUEST']._serialized_end=2804
  _globals['_LISTFILESRESPONSE']._serialized_start=2806
  _globals['_LISTFILESRESPONSE']._serialized_end=2840
  _globals['_FINDCOMMITREQUEST']._serialized_start=2842
  _globals['_FINDCOMMITREQUEST']._serialized_end=2943
  _globals['_FINDCOMMITRESPONSE']._serialized_start=2945
  _globals['_FINDCOMMITRESPONSE']._serialized_end=3000
  _globals['_LISTCOMMITSBYOIDREQUEST']._serialized_start=3002
  _globals['_LISTCOMMITSBYOIDREQUEST']._serialized_end=3086
  _globals['_LISTCOMMITSBYOIDRESPONSE']._serialized_start=3088
  _globals['_LISTCOMMITSBYOIDRESPONSE']._serialized_end=3150
  _globals['_LISTCOMMITSBYREFNAMEREQUEST']._serialized_start=3152
  _globals['_LISTCOMMITSBYREFNAMEREQUEST']._serialized_end=3246
  _globals['_LISTCOMMITSBYREFNAMERESPONSE']._serialized_start=3249
  _globals['_LISTCOMMITSBYREFNAMERESPONSE']._serialized_end=3426
  _globals['_LISTCOMMITSBYREFNAMERESPONSE_COMMITFORREF']._serialized_start=3353
  _globals['_LISTCOMMITSBYREFNAMERESPONSE_COMMITFORREF']._serialized_end=3420
  _globals['_FINDALLCOMMITSREQUEST']._serialized_start=3429
  _globals['_FINDALLCOMMITSREQUEST']._serialized_end=3640
  _globals['_FINDALLCOMMITSREQUEST_ORDER']._serialized_start=529
  _globals['_FINDALLCOMMITSREQUEST_ORDER']._serialized_end=566
  _globals['_FINDALLCOMMITSRESPONSE']._serialized_start=3642
  _globals['_FINDALLCOMMITSRESPONSE']._serialized_end=3702
  _globals['_FINDCOMMITSREQUEST']._serialized_start=3705
  _globals['_FINDCOMMITSREQUEST']._serialized_end=4237
  _globals['_FINDCOMMITSREQUEST_ORDER']._serialized_start=529
  _globals['_FINDCOMMITSREQUEST_ORDER']._serialized_end=556
  _globals['_FINDCOMMITSRESPONSE']._serialized_start=4239
  _globals['_FINDCOMMITSRESPONSE']._serialized_end=4296
  _globals['_FINDCOMMITSERROR']._serialized_start=4299
  _globals['_FINDCOMMITSERROR']._serialized_end=4485
  _globals['_COMMITLANGUAGESREQUEST']._serialized_start=4487
  _globals['_COMMITLANGUAGESREQUEST']._serialized_end=4575
  _globals['_COMMITLANGUAGESRESPONSE']._serialized_start=4578
  _globals['_COMMITLANGUAGESRESPONSE']._serialized_end=4753
  _globals['_COMMITLANGUAGESRESPONSE_LANGUAGE']._serialized_start=4666
  _globals['_COMMITLANGUAGESRESPONSE_LANGUAGE']._serialized_end=4753
  _globals['_RAWBLAMEREQUEST']._serialized_start=4755
  _globals['_RAWBLAMEREQUEST']._serialized_end=4865
  _globals['_RAWBLAMERESPONSE']._serialized_start=4867
  _globals['_RAWBLAMERESPONSE']._serialized_end=4899
  _globals['_RAWBLAMEERROR']._serialized_start=4902
  _globals['_RAWBLAMEERROR']._serialized_end=5083
  _globals['_RAWBLAMEERROR_OUTOFRANGEERROR']._serialized_start=5035
  _globals['_RAWBLAMEERROR_OUTOFRANGEERROR']._serialized_end=5074
  _globals['_LASTCOMMITFORPATHREQUEST']._serialized_start=5086
  _globals['_LASTCOMMITFORPATHREQUEST']._serialized_end=5263
  _globals['_LASTCOMMITFORPATHRESPONSE']._serialized_start=5265
  _globals['_LASTCOMMITFORPATHRESPONSE']._serialized_end=5327
  _globals['_LISTLASTCOMMITSFORTREEREQUEST']._serialized_start=5330
  _globals['_LISTLASTCOMMITSFORTREEREQUEST']._serialized_end=5547
  _globals['_LISTLASTCOMMITSFORTREERESPONSE']._serialized_start=5550
  _globals['_LISTLASTCOMMITSFORTREERESPONSE']._serialized_end=5737
  _globals['_LISTLASTCOMMITSFORTREERESPONSE_COMMITFORTREE']._serialized_start=5655
  _globals['_LISTLASTCOMMITSFORTREERESPONSE_COMMITFORTREE']._serialized_end=5737
  _globals['_COMMITSBYMESSAGEREQUEST']._serialized_start=5740
  _globals['_COMMITSBYMESSAGEREQUEST']._serialized_end=5936
  _globals['_COMMITSBYMESSAGERESPONSE']._serialized_start=5938
  _globals['_COMMITSBYMESSAGERESPONSE']._serialized_end=6000
  _globals['_FILTERSHASWITHSIGNATURESREQUEST']._serialized_start=6002
  _globals['_FILTERSHASWITHSIGNATURESREQUEST']._serialized_end=6095
  _globals['_FILTERSHASWITHSIGNATURESRESPONSE']._serialized_start=6097
  _globals['_FILTERSHASWITHSIGNATURESRESPONSE']._serialized_end=6145
  _globals['_GETCOMMITSIGNATURESREQUEST']._serialized_start=6147
  _globals['_GETCOMMITSIGNATURESREQUEST']._serialized_end=6241
  _globals['_GETCOMMITSIGNATURESRESPONSE']._serialized_start=6244
  _globals['_GETCOMMITSIGNATURESRESPONSE']._serialized_end=6462
  _globals['_GETCOMMITSIGNATURESRESPONSE_SIGNER']._serialized_start=6394
  _globals['_GETCOMMITSIGNATURESRESPONSE_SIGNER']._serialized_end=6462
  _globals['_GETCOMMITMESSAGESREQUEST']._serialized_start=6464
  _globals['_GETCOMMITMESSAGESREQUEST']._serialized_end=6556
  _globals['_GETCOMMITMESSAGESRESPONSE']._serialized_start=6558
  _globals['_GETCOMMITMESSAGESRESPONSE']._serialized_end=6621
  _globals['_CHECKOBJECTSEXISTREQUEST']._serialized_start=6623
  _globals['_CHECKOBJECTSEXISTREQUEST']._serialized_end=6714
  _globals['_CHECKOBJECTSEXISTRESPONSE']._serialized_start=6717
  _globals['_CHECKOBJECTSEXISTRESPONSE']._serialized_end=6867
  _globals['_CHECKOBJECTSEXISTRESPONSE_REVISIONEXISTENCE']._serialized_start=6818
  _globals['_CHECKOBJECTSEXISTRESPONSE_REVISIONEXISTENCE']._serialized_end=6867
  _globals['_COMMITSERVICE']._serialized_start=6870
  _globals['_COMMITSERVICE']._serialized_end=9031
# @@protoc_insertion_point(module_scope)



# Copyright 2020-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from .git import (
    OBJECT_MODE_EXECUTABLE,
    OBJECT_MODE_LINK,
    OBJECT_MODE_NON_EXECUTABLE,
)


def git_perms(filectx):
    """Return Git representation for file permissions (aka mode).

    Reference: https://git-scm.com/book/en/v2/Git-Internals-Git-Objects
    """
    if filectx.islink():
        return OBJECT_MODE_LINK
    elif filectx.isexec():
        return OBJECT_MODE_EXECUTABLE
    else:
        return OBJECT_MODE_NON_EXECUTABLE



# coding: utf-8
# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
import os
from pathlib import Path
from urllib.parse import urlparse

from mercurial import (
    hg,
)

from .ssh import client_command_str

ALLOWED_URL_SCHEMES = ('http', 'https', 'ssh', 'file')


class URLParseError(ValueError):
    """Arguments: (the invalid URL, exception)."""


class InvalidURLScheme(ValueError):
    """Arguments: (the invalid scheme, list of admissible schemes)."""


class FileURLOutsidePath(ValueError):
    """Arguments: incoming URL, allowed path."""


class PeerInitException(RuntimeError):
    """Argument is the underlying exception."""


def init_peer(repo, **kw):
    """Encapsulation of `hg.peer` instantiation with a dedicated exception.


    The dedication exception h with differentiated handling by callers.
    """
    try:
        return hg.peer(repo, **kw)
    except Exception as exc:
        raise PeerInitException(str(exc))


@contextmanager
def hg_remote_peer(repo, peer_description, storage_path,
                   allowed_url_schemes=ALLOWED_URL_SCHEMES):
    """Creation of remote peer from gRPC requests.

    :param peer_description: object having the same attributes as
       the `MercurialPeer` gRPC message

    Because this context manager will be used to create the peer, and then
    to perform some operation on it, the exceptions that can happen on enter
    are catched and reraised as :class:`PeerInitException` so that the
    caller can tell them apart from those raised while using the peer.
    """
    remote_url = peer_description.url
    remote_url_bytes = remote_url.encode('utf-8')
    opts = {}
    try:
        parsed_url = urlparse(remote_url)
    except Exception as exc:
        raise URLParseError(remote_url, exc)

    scheme = parsed_url.scheme or 'file'
    if scheme not in allowed_url_schemes:
        raise InvalidURLScheme(scheme, allowed_url_schemes)

    if scheme == 'file':
        storage_path = os.path.realpath(storage_path)

        # If ever there was HGitaly running on Windows, Pathlib join
        # operator works as one would expect:
        # - the right operand can contain forward slashes,
        #   they get converted to the os separator
        # - a leading forward slash (absolute path of URIs) anchors to
        #   the root of the disk, e.g., `C:\`
        target_path = os.path.realpath(Path(storage_path) / parsed_url.path)
        if not target_path.startswith(storage_path + os.path.sep):
            raise FileURLOutsidePath(remote_url, storage_path)

    if scheme == 'ssh':
        with client_command_str(peer_description.ssh_key,
                                peer_description.ssh_known_hosts) as ssh:
            opts[b'ssh'] = ssh.encode('ascii')
            remote_cmd = peer_description.ssh_remote_command
            if remote_cmd:
                opts[b'remotecmd'] = remote_cmd
            yield init_peer(repo, opts=opts, path=remote_url_bytes)
    else:
        yield init_peer(repo, opts=opts, path=remote_url_bytes)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Quoting from Gitaly 13.4:
# //  WriteBufferSize is the largest []byte that Write() will pass
# //  to its underlying send function. This value can be changed
# //  at runtime using the GITALY_STREAMIO_WRITE_BUFFER_SIZE environment
# //  variable.
#
# var WriteBufferSize = 128 * 1024
#
# As of GitLab 13.4, the environment variable is parsed with
# `strconv.ParseInt(value, 0, 32)`.
# Quoting https://golang.org/pkg/strconv/#ParseInt:
#     If the base argument is 0, the true base is implied by
#     the string's prefix: 2 for "0b", 8 for "0" or "0o", 16 for "0x",
#     and 10 otherwise. Also, for argument base 0 only,
#     underscore characters are permitted as defined by the
#     Go syntax for integer literals.
import contextlib
import os
import tempfile


def concat_resplit(in_chunks, out_chunks_size):
    """Generator that aggregate incoming chunks of bytes and yield chunks with
    the wished size.

    in_chunks: an iterator of chunk of bytes of arbitrary sizes
    out_chunks_size: size of chunks to be yield, except last one
    """
    data = b''
    for chunk in in_chunks:
        data += chunk
        while len(data) > out_chunks_size:
            yield data[:out_chunks_size]
            data = data[out_chunks_size:]
    yield data


def split_batches(in_bytes, out_bytes_size):
    """Generator that yield in_bytes in fixed size batches.

    in_bytes: bytes of arbitrary size
    out_bytes_size: size of chunks to be yield, except last one
    """
    while len(in_bytes) > out_bytes_size:
        yield in_bytes[:out_bytes_size]
        in_bytes = in_bytes[out_bytes_size:]
    yield in_bytes


def iter_boolean_lookahead(itr):
    """Generator that yield tuple (value, is_last_value)."""
    try:
        prev = next(itr)
    except StopIteration:
        return

    for value in itr:
        yield prev, False
        prev = value
    yield prev, True


def aggregate_flush_batch(itr, custom_size_func, out_batch_size):
    """Generator that yields `itr` values in batch after the size cross
    a certain limit, where size of each value is calculated using a
    `custom_size_func`.

    itr: an iterator of values that to be yield in batch
    custom_size_func: a func to calculate size of each value
    out_batch_size: threshold of a batch size
    """
    batch = []
    batch_size = 0
    for val in itr:
        batch.append(val)
        batch_size += custom_size_func(val)
        if batch_size > out_batch_size:
            yield batch
            batch = []
            batch_size = 0
    yield batch


def chunked_limit(itr, limit):
    """Wrap an iterator of lists to limit the total number of yielded items.

    The iterator is not called at all once the limit has been reached.
    This can be a performance advantage.

    :param limit: if ``None``, no limiting occurs, otherwise the total
      number of wished items
    """
    for chunk in itr:
        if limit is not None:
            if not limit:
                break
            chunk = chunk[:limit]
            limit = limit - len(chunk)
        yield chunk


def parse_int(s):
    """Parse integer string representations, as Golangs `strconf.ParseInt`

    # TODO check at least octal and hex syntaxes
    >>> parse_int('10')
    10
    """
    return int(s)


def env_write_buffer_size():
    str_val = os.environ.get('GITALY_STREAMIO_WRITE_BUFFER_SIZE')
    if not str_val:
        return 128 * 1024
    return parse_int(str_val)


WRITE_BUFFER_SIZE = env_write_buffer_size()
"""In Gitaly, there is a general binary chunking system,
used in many places, notably all binary content producers (diff, blame,
blob, archive...)

Extract from streamio/stream.go:

// WriteBufferSize is the largest []byte that Write() will pass to its
// underlying send function. This value can be changed at runtime using
// the GITALY_STREAMIO_WRITE_BUFFER_SIZE environment variable.
var WriteBufferSize = 128 * 1024
"""


@contextlib.contextmanager
def streaming_request_tempfile_extract(requests, context,
                                       data_attr='data',
                                       first_request_handler=None):
    """Extract data from a streaming request.

    It is a a common Gitaly pattern for streaming requests with binary
    data to bear other information in the first request message only.

    This context manager extracts all the binary data in a
    :class:`NamedTemporaryFile` and calls `first_request_handler` to
    perform actions from the first request message.

    The provided Python context is the (first_result, temporary file) pair,
    where `first_result` is the return value of `first_request_handler`.
    """
    with tempfile.NamedTemporaryFile(mode='wb+',
                                     buffering=WRITE_BUFFER_SIZE) as tmpf:
        first_request = True
        first_result = None
        for req in requests:
            if first_request and first_request_handler is not None:
                first_result = first_request_handler(req, context)
                first_request = False
            tmpf.write(req.data)
        tmpf.seek(0)

        yield first_result, tmpf


def slice_binary_file(path, nb_chunks):
    """Generator for the content of binary file at path.

    Suboptimal implementation currently used in tests only (loads
    file content in RAM).
    Must be improved for use in main code (not so likely because main code
    would be more interested in a fixed chunk size).
    """
    with open(path, 'rb') as fobj:
        data = fobj.read()

    chunk_size = len(data) // nb_chunks
    for i in range(0, nb_chunks):
        yield data[i * chunk_size:(i + 1) * chunk_size]



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""gRPC interceptors for HGitaly.

TODO This could be placed in a subpackage for all HGitaly code relying on
Protobuf/gRPC concepts rather than sitting alongseide services without
being a service implementations. We would thus have
`hgitaly.grpc.interceptors`, `hgitaly.grpc.service.ref`, perhaps even
`hgitaly.grpc.message`.
"""
import itertools
import logging
import time
from typing import (
    Any,
    Callable
)

import grpc
from grpc_interceptor import ServerInterceptor

from .. import message
from ..logging import LoggerAdapter


base_logger = logging.getLogger(__name__)


def is_streaming(message) -> bool:
    """Tell whether a gRPC message is streaming (iterable).

    Note that messages aren't plain strings nor Python collections.
    """
    return hasattr(message, '__iter__')


def stream_with_time_log(response, start, logger, extra_log):
    yield from response

    extra_log['elapsed_seconds'] = time.time() - start
    logger.info("Finished RPC", extra=extra_log)


class RequestLoggerInterceptor(ServerInterceptor):
    """Log every request.

    Mostly taken from the example in grpc-interceptor documentation.

    In case of streaming requests, only the first one of the stream gets
    logged. With Gitaly convention, this is usually the only one carrying
    interesting metadata.
    """

    def intercept(
            self,
            method: Callable,
            request: Any,
            context: grpc.ServicerContext,
            method_name: str,
    ) -> Any:
        if is_streaming(request):
            first = next(iter(request))
            request = itertools.chain([first], request)
        else:
            first = request

        # of course it would be great to change the logger depending
        # on the service, but with (H)Gitaly naming conventions, the class
        # name of the request contains all the information.
        start = time.time()
        extra_log = dict(request=message.Logging(first))
        logger = LoggerAdapter(base_logger, context)
        logger.info("Starting to process RPC", extra=extra_log)

        response = method(request, context)

        if is_streaming(response):
            return stream_with_time_log(response, start, logger, extra_log)
        else:
            extra_log['elapsed_seconds'] = time.time() - start
            logger.info("Finished RPC", extra=extra_log)
            return response



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Tests for the GitalyServer helper

These don't involve HGitaly and not even Mercurial. The purpose is to
ensure that the foundation of all Gitaly comparison tests keeps working.
"""
import pytest

from hgitaly.stub.repository_pb2 import (
    CreateRepositoryRequest,
    RepositoryExistsRequest,
)
from hgitaly.stub.shared_pb2 import Repository
from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub

from . import skip_comparison_tests
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip


def test_gitaly_channel(gitaly_channel, server_repos_root):
    channel = gitaly_channel
    repo_stub = RepositoryServiceStub(channel)

    repo = Repository(relative_path='foo.git', storage_name='default')
    repo_stub.CreateRepository(CreateRepositoryRequest(repository=repo))

    # TODO should have a test util for that:
    # join(server_repos_root, repo)
    git_path = server_repos_root / 'default' / 'foo.git'
    assert git_path.is_dir()
    assert (git_path / 'HEAD').exists

    assert repo_stub.RepositoryExists(
        RepositoryExistsRequest(repository=repo)
    ).exists
